/* ****************************************************************************
* Программа: список студентов
* Модуль:    работа с описаниями студенческих групп
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// заголовочный файл нужен для функции strlen()
#include <string.h>

// для поддержки широких символов
#include <wchar.h>

#include "students.h"


// Ключевое слово static означает, что данная функция доступна только для
// других функций, содержащихся в этом же программном модуле. Поэтому мы
// сможем использовать одинаковые имена функций в разных программных модулях.
// Это может быть удобно, когда такие функции находятся в различных модулях,
// но выполняют одинаковые операции, например, ввод данных.
static int add( void );
static int find( void );
static int list( void );
static int delete( void );
static int update( void );


// объявим переменную типа GROUP
// все поля этой структуры при инициализации получают значения 0, т. е. будут
// состоять только из символов '\0' (это можно увидеть с помощью отладчика
// или если открыть файл groups.db в текстовом редакторе, который "умеет"
// показывать ASCII-код того символа, на котором стоит курсор)
static GROUP grp;

/* ****************************************************************************
* формирование меню для работы с описаниями студенческих групп
* ****************************************************************************/
void groups( void )
{
  char answer;     // ответ пользователя при выборе пункта меню

  printf( "\nОписания студенческих групп\n" );

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = '1';

  while ( answer == '1' || answer == '2' || answer == '3' || 
          answer == '4' || answer == '5' )
  {
    printf( "\n" );              // пустая строка
    printf( "1. Перечень студенческих групп\n" );
    printf( "2. Ввод нового описания группы\n" );
    printf( "3. Корректировка описания группы\n" );
    printf( "4. Удаление описания группы\n" );
    printf( "5. Поиск описания группы\n" );
    printf( "6. Выход в главное меню программы\n\n" );

    printf( "Выберите действие: " );
    scanf( "%c", &answer );      // ответ получим в виде символа, а не числа
    printf( "\n" );              // пустая строка

    // очистим буфер ввода
    while ( getchar() != '\n' )
      continue;

    if ( answer == '1' )
    {
      list();
    }
    else if ( answer == '2' )
    {
      add();
    }
    else if ( answer == '3' )
    {
      update();
    }
    else if ( answer == '4' )
    {
      delete();
    }
    else if ( answer == '5' )
    {
      find();
    }
    else if ( answer == '6' )
    {
      return;
    }
    else    // выбран какой-то другой символ
    {
//      clrscr();      // очистим экран
      printf( "Укажите правильный номер пункта меню\n" );

      // присвоим переменной answer такое значение, которое позволит нам 
      // остаться в цикле
      answer = '1';
    }
  }
  return;
}


/* ****************************************************************************
* ввод нового описания студенческой группы
* ****************************************************************************/
static int add( void )
{
#ifdef UTF8
  wchar_t answer;  // ответ пользователя при вводе данных
#else
  char answer;     // ответ пользователя при вводе данных
#endif

  int written;     // счетчик числа блоков данных, успешно записанных в файл

  // файл для записи описаний групп
  FILE *out;

  // откроем двоичный файл описаний групп в режиме добавления записей
  // в конец файла (дадим нашему файлу расширение "db", что означает database)
  out = fopen( "groups.db", "ab" );

  // если файл открыть (создать) не удалось, то указатель получит значение NULL
  if ( out == NULL )
  {
    fprintf( stderr, "Не могу открыть файл groups.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Ввод нового описания группы\n\n" );

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = YES;

  while ( answer == YES )
  {
    // будем вводить каждое поле на отдельной строке
    printf( "Шифр группы (не более %d символов без пробелов): ",
            GROUP_CODE_LEN - 1 );
    scanf( "%s", grp.group_code );

    // очистим буфер ввода
    while ( getchar() != '\n' )
      continue;

    printf( "\nОписание группы (не более %d символов): ", GROUP_DESCR_LEN - 1 );
    gets( grp.group_descr );

    // в поле "признак удаленной записи" запишем значение '\0', т. е. эта
    // запись при вводе в нашу скромную базу данных, естественно, не будет
    // помечена как удаленная
    grp.deleted = '\0';

    // в данном случае буфер ввода очищать не нужно, т. к. функция gets()
    // сама забирает символ '\n' из буфера
    // попробуйте раскомментировать следующие две строки текста и посмотрите,
    // что изменится в поведении программы в данном месте
    // while ( getchar() != '\n' )
    //   continue;

    // запишем введенные данные в файл groups.db
    written = fwrite( &grp, sizeof( grp ), 1, out );

    if ( written != 1 )
      printf( "Ошибка при записи данных в файл groups.db\n" );

    // спросим у пользователя, хочет ли он продолжить ввод, причем, будем
    // настаивать на том, чтобы пользователь для продолжения ввода ввел "д",
    // а для завершения ввода ввел не какую-нибудь другую букву, а именно "н"
    // (чтобы отказаться от продолжения ввода осмысленно, а не поспешно)

    // присвоим переменной answer такое значение, которое позволит нам войти
    // внутрь цикла
    answer = '\0';

    while ( answer != YES && answer != NO )
    {
      printf( "\nПродолжить ввод данных (д/н) ?: " );
      scanf( FORMAT, &answer );

      // очистим буфер ввода
      while ( getchar() != '\n' )
        continue;

    }
  }  
  
  // закроем файл
  fclose( out );

  return;
}


/* ****************************************************************************
* удаление одного описания студенческой группы из файла groups.db
* ****************************************************************************/
static int delete( void )
{
  int  written;    // счетчик числа блоков данных, успешно записанных в файл

  // шифр студенческой группы в файле
  char grp_code[ GROUP_CODE_LEN ];

  char deleted_mark = '*';  // метка удаленной записи
  int found = 0;            // признак того, что запись в файле найдена
 
  // файл для хранения описаний групп
  FILE *groups;

  // откроем файл в режиме чтения и записи 
  groups = fopen( "groups.db", "r+b" );

  // проверяем успешность открытия файла
  if ( groups == NULL )
  {
    fprintf( stderr, "Не могу открыть файл groups.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Удаление описания группы\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  scanf( "%s", grp_code );

  // очистим буфер ввода
  while ( getchar() != '\n' )
    continue;

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей описание одной группы
  while ( fread( &grp, sizeof( grp ), 1, groups ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( grp.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенный пользователем шифр группы совпадает с тем, который
    // содержится в прочитанной записи из файла
    if ( strcmp( grp_code, grp.group_code ) == 0 )
    {
      found = 1;    // запись найдена -- присвоим флагу ненулевое значение

      // переместим указатель позиции в файле к началу этой записи (структуры)
      // параметр SEEK_CUR означает, что перемещение указателя позиции
      // выполняется от ТЕКУЩЕЙ позиции в файле; знак "-" означает перемещение
      // в сторону НАЧАЛА файла
      fseek( groups, -( sizeof( grp ) ), SEEK_CUR );

      // запишем метку удаления в найденную запись в файле groups.db
      written = fwrite( &deleted_mark, 1, 1, groups );

      if ( written == 1 )
        printf( "Запись удалена\n" );
      else
        printf( "Ошибка при записи данных в файл groups.db\n" );

      break;      // выходим из цикла, т. к. запись найдена и обработана
    }
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенный пользователем шифр группы в файле не найден
    printf( "Неверный шифр группы\n" );

  // закроем файл
  fclose( groups );

  return 0;        // успешное завершение процедуры
}


/* ****************************************************************************
* поиск описания студенческой группы в файле groups.db
* ****************************************************************************/
static int find( void )
{
  // шифр студенческой группы в файле
  char grp_code[ GROUP_CODE_LEN ];

  int found = 0;            // признак того, что запись в файле найдена
 
  // файл для хранения описаний групп
  FILE *groups;

  // откроем файл в режиме чтения
  groups = fopen( "groups.db", "rb" );

  // проверяем успешность открытия файла
  if ( groups == NULL )
  {
    fprintf( stderr, "Не могу открыть файл groups.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Поиск описания группы\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  scanf( "%s", grp_code );

  // очистим буфер ввода
  while ( getchar() != '\n' )
    continue;

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей описание одной группы
  while ( fread( &grp, sizeof( grp ), 1, groups ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( grp.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенный пользователем шифр группы совпадает с тем, который
    // содержится в прочитанной записи из файла
    if ( strcmp( grp_code, grp.group_code ) == 0 )
    {
      found = 1;    // запись найдена -- присвоим флагу ненулевое значение

      // выведем найденную запись одной строкой
      printf( "%s %s\n", grp.group_code, grp.group_descr );

      break;        // выходим из цикла, т. к. запись найдена и обработана
    }
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенный пользователем шифр группы в файле не найден
    printf( "Группы с шифром %s нет в файле groups.db\n", grp_code );

  // закроем файл
  fclose( groups );

  return 0;        // успешное завершение процедуры
}


/* ****************************************************************************
* вывод списка описаний студенческих групп
* ****************************************************************************/
static int list( void )
{
  int written;     // счетчик числа блоков данных, успешно записанных в файл
  int i;           // счетчик числа выведенных записей

  // файл для хранения описаний групп
  FILE *in;

  // откроем теперь файл в режиме чтения и выведем содержимое файла
  in = fopen( "groups.db", "rb" );

  // проверяем успешность открытия файла
  if ( in == NULL )
  {
    fprintf( stderr, "Не могу открыть файл groups.db\n" );
    return 1;      // неуспешное завершение процедуры
  }

  printf( "Перечень студенческих групп\n\n" );

  // будем в цикле читать из файла groups.db структуру за структурой
  // (т. е. запись за записью)

  i = 1;   // будем нумеровать выводимые записи о группах

  while ( fread( &grp, sizeof( grp ), 1, in ) == 1 )
  {
    // если запись в файле не помечена как удаленная (т. е. значение поля
    // deleted так и осталось нулевым)...
    if ( ! grp.deleted )
    {
      // ... то выведем эту запись одной строкой
      printf( "%d. %s %s\n", i, grp.group_code, grp.group_descr );
      i++;
    }
  }

  // закроем файл
  fclose( in );

  return 0;
}


/* ****************************************************************************
* корректировка одного описания студенческой группы в файле groups.db
* ****************************************************************************/
static int update( void )
{
  int  written;    // счетчик числа блоков данных, успешно записанных в файл

  char grp_code[ GROUP_CODE_LEN ];      // шифр студенческой группы
  char grp_descr[ GROUP_DESCR_LEN ];    // описание группы

  int found = 0;            // признак того, что запись в файле найдена
  int updated = 0;          // признак того, что запись в файле обновлена

  // файл для хранения описаний групп
  FILE *groups;

  // откроем файл в режиме чтения и записи 
  groups = fopen( "groups.db", "r+b" );

  // проверяем успешность открытия файла
  if ( groups == NULL )
  {
    fprintf( stderr, "Не могу открыть файл groups.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Корректировка описания группы\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  scanf( "%s", grp_code );

  // очистим буфер ввода
  while ( getchar() != '\n' )
    continue;

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей описание одной группы
  while ( fread( &grp, sizeof( grp ), 1, groups ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( grp.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенный пользователем шифр группы НЕ совпадает с тем, который
    // содержится в прочитанной записи из файла
    if ( strcmp( grp_code, grp.group_code ) != 0 )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    found = 1;     // запись найдена -- присвоим флагу ненулевое значение

    // выведем на экран текущие значения полей записи
    printf( "%s   %s\n\n", grp.group_code, grp.group_descr );

    // попросим пользователя ввести данные
    printf( "Введите новые значения полей описания студенческой группы.\n" \
            "Если необходимо оставить прежнее значение поля, " \
            "просто нажмите клавишу Enter\n\n");

    // будем вводить каждое поле на отдельной строке
    printf( "Шифр группы (не более %d символов без пробелов): ",
            GROUP_CODE_LEN - 1 );
    gets( grp_code );

    printf( "\nОписание группы (не более %d символов): ", GROUP_DESCR_LEN - 1 );
    gets( grp_descr );

    printf( "\n" );   // пустая строка на экране

    // в данном случае буфер ввода очищать не нужно, т. к. функция gets()
    // сама забирает символ '\n' из буфера
    // попробуйте раскомментировать следующие две строки текста и посмотрите,
    // что изменится в поведении программы в данном месте
    // while ( getchar() != '\n' )
    //   continue;

    // если ввели новое значение в поле "Шифр группы"
    if ( strlen( grp_code ) != 0 )
    {
      // скопируем новое значение в структуру 
      strcpy( grp.group_code, grp_code );
      updated = 1;                       // признак -- запись обновлена
    }

    // если ввели новое значение в поле "Описание группы"
    if ( strlen( grp_descr ) != 0 )
    {
      // скопируем новое значение в структуру 
      strcpy( grp.group_descr, grp_descr );
      updated = 1;                        // признак -- запись обновлена
    }

    // если было введено новое значение хотя бы для одного из полей
    if ( updated )   // это условие можно записать так: if ( updated == 1 )
    {
      // переместим указатель позиции в файле к началу этой записи (структуры)
      // параметр SEEK_CUR означает, что перемещение указателя позиции
      // выполняется от ТЕКУЩЕЙ позиции в файле; знак "-" означает перемещение
      // в сторону НАЧАЛА файла
      fseek( groups, -( sizeof( grp ) ), SEEK_CUR );

      // запишем введенные данные в файл groups.db
      written = fwrite( &grp, sizeof( grp ), 1, groups );

      if ( written == 1 )
        printf( "Запись обновлена\n" );
      else
        printf( "Ошибка при записи данных в файл groups.db\n" );
    }
    else         // т. е. updated == 0
    {
      printf( "Запись не обновлена, т. к. не введены новые значения полей\n" );
    }
    
    break;      // выходим из цикла, т. к. запись найдена и обработана
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенный пользователем шифр группы в файле не найден
    printf( "Неверный шифр группы\n" );

  // закроем файл
  fclose( groups );

  return 0;        // успешное завершение процедуры
}
