/* ****************************************************************************
* Программа: список студентов
* Модуль:    работа со списками студенческих групп
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// заголовочный файл нужен для функции strlen()
#include <string.h>

// для поддержки широких символов
#include <wchar.h>

#include "students.h"


// Ключевое слово static означает, что данная функция доступна только для
// других функций, содержащихся в этом же программном модуле. Поэтому мы
// сможем использовать одинаковые имена функций в разных программных модулях.
// Это может быть удобно, когда такие функции находятся в различных модулях,
// но выполняют одинаковые операции, например, ввод данных.
static int add( void );
static int find( void );
static int list( void );
static int delete( void );
static int update( void );


// объявим переменную типа STUDENT
// все поля этой структуры при инициализации получают значения 0, т. е. будут
// состоять только из символов '\0' (это можно увидеть с помощью отладчика
// или если открыть файл students.db в текстовом редакторе, который "умеет"
// показывать ASCII-код того символа, на котором стоит курсор)

// регистр символов имеет значение, поэтому можно использовать такие имена
static STUDENT student;


/* ****************************************************************************
* формирование меню для работы со списками студенческих групп
* ****************************************************************************/
void lists( void )
{
  char answer;     // ответ пользователя при выборе пункта меню

  printf( "\nСписки студенческих групп\n" );

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = '1';

  while ( answer == '1' || answer == '2' || answer == '3' || 
          answer == '4' || answer == '5' )
  {
    printf( "\n" );              // пустая строка
    printf( "1. Списки студенческих групп\n" );
    printf( "2. Ввод данных студента\n" );
    printf( "3. Корректировка данных студента\n" );
    printf( "4. Удаление данных студента\n" );
    printf( "5. Поиск данных студента\n" );
    printf( "6. Выход в главное меню программы\n\n" );

    printf( "Выберите действие: " );
    scanf( "%c", &answer );      // ответ получим в виде символа, а не числа
    printf( "\n" );              // пустая строка

    // очистим буфер ввода
    while ( getchar() != '\n' )
      continue;

    if ( answer == '1' )
    {
      list();
    }
    else if ( answer == '2' )
    {
      add();
    }
    else if ( answer == '3' )
    {
      update();
    }
    else if ( answer == '4' )
    {
      delete();
    }
    else if ( answer == '5' )
    {
      find();
    }
    else if ( answer == '6' )
    {
      return;
    }
    else    // выбран какой-то другой символ
    {
      printf( "Укажите правильный номер пункта меню\n" );

      // присвоим переменной answer такое значение, которое позволит нам 
      // остаться в цикле
      answer = '1';
    }
  }
  return;
}


/* ****************************************************************************
* ввод данных студента
* ****************************************************************************/
static int add( void )
{
  // это счетчик количества записей о студентах, введенных за время текущего
  // сеанса работы с програмой
  // переменная с ключевым словом static сохраняет свое значение и между
  // вызовами той функции, в которой она определена
  static int count_students = 0;
  
#ifdef UTF8
  wchar_t answer;  // ответ пользователя при вводе данных
#else
  char answer;     // ответ пользователя при вводе данных
#endif

  int written;     // счетчик числа блоков данных, успешно записанных в файл

  // файл для хранения списков студентов
  FILE *out;

  // откроем двоичный файл списков студентов в режиме добавления записей
  // в конец файла (дадим нашему файлу расширение "db", что означает database)
  out = fopen( "students.db", "ab" );

  // если файл открыть (создать) не удалось, то указатель получит значение NULL
  if ( out == NULL )
  {
    fprintf( stderr, "Не могу открыть файл students.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Ввод данных студента\n\n" );

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = YES;

  while ( answer == YES )
  {
    // будем вводить каждое поле на отдельной строке
    printf( "Шифр группы (не более %d символов без пробелов): ",
            GROUP_CODE_LEN - 1 );
    gets( student.group_code );

    printf( "\nФамилия (не более %d символов): ", NAME_LEN - 1 );
    gets( student.last_name );

    printf( "\nИмя (не более %d символов): ", NAME_LEN - 1 );
    gets( student.first_name );

    printf( "\nРост (в сантиметрах): " );
    scanf( "%d", &student.height );

    printf( "\nВес (в килограммах) " \
            "(десятичную часть отделяйте символом \".\"): " );
    scanf( "%f", &student.weight );

    // очистим буфер ввода
    while ( getchar() != '\n' )
      continue;
 
    // в поле "признак удаленной записи" запишем значение '\0', т. е. эта
    // запись при вводе в нашу скромную базу данных, естественно, не будет
    // помечена как удаленная
    student.deleted = '\0';

    // запишем введенные данные в файл students.db
    written = fwrite( &student, sizeof( student ), 1, out );

    if ( written != 1 )
      printf( "Ошибка при записи данных в файл students.db\n" );
    else
      printf( "В этом сеансе работы с программой вы ввели %d записей "
              "о студентах\n", ++count_students );
      
    // спросим у пользователя, хочет ли он продолжить ввод, причем, будем
    // настаивать на том, чтобы пользователь для продолжения ввода ввел "д",
    // а для завершения ввода ввел не какую-нибудь другую букву, а именно "н"
    // (чтобы отказаться от продолжения ввода осмысленно, а не поспешно)

    // присвоим переменной answer такое значение, которое позволит нам войти
    // внутрь цикла
    answer = '\0';

    while ( answer != YES && answer != NO )
    {
      printf( "\nПродолжить ввод данных (д/н) ?: " );
      scanf( FORMAT, &answer );

      // очистим буфер ввода
      while ( getchar() != '\n' )
        continue;

    }
  }  
  
  // закроем файл
  fclose( out );

  return;
}


/* ****************************************************************************
* удаление данных одного студента из файла students.db
* ****************************************************************************/
static int delete( void )
{
  int  written;    // счетчик числа блоков данных, успешно записанных в файл

  // шифр студенческой группы в файле
  char grp_code[ GROUP_CODE_LEN ];

  // фамилия студента в файле
  char last_name[ NAME_LEN ];

  char deleted_mark = '*';  // метка удаленной записи
  int found = 0;            // признак того, что запись в файле найдена
 
  // файл для хранения списков студентов
  FILE *students;

  // откроем файл в режиме чтения и записи 
  students = fopen( "students.db", "r+b" );

  // проверяем успешность открытия файла
  if ( students == NULL )
  {
    fprintf( stderr, "Не могу открыть файл students.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Удаление данных студента\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  gets( grp_code );

  printf( "\nВведите фамилию студента (не более %d символов): ", NAME_LEN - 1 );
  gets( last_name );

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей данные одного студента
  while ( fread( &student, sizeof( student ), 1, students ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( student.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенные пользователем шифр группы и фамилия студента совпадают
    // с теми, которые содержатся в прочитанной записи из файла
    if ( strcmp( grp_code, student.group_code ) == 0 && 
         strcmp( last_name, student.last_name ) == 0
       )
    {
      found = 1;    // запись найдена -- присвоим флагу ненулевое значение

      // переместим указатель позиции в файле к началу этой записи (структуры)
      // параметр SEEK_CUR означает, что перемещение указателя позиции
      // выполняется от ТЕКУЩЕЙ позиции в файле; знак "-" означает перемещение
      // в сторону НАЧАЛА файла
      fseek( students, -( sizeof( student ) ), SEEK_CUR );

      // запишем метку удаления в найденную запись в файле students.db
      written = fwrite( &deleted_mark, 1, 1, students );

      if ( written == 1 )
        printf( "Запись удалена\n" );
      else
        printf( "Ошибка при записи данных в файл students.db\n" );

      break;      // выходим из цикла, т. к. запись найдена и обработана
    }
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенные пользователем ключевые данные в файле не найдены
    printf( "Неверный шифр группы или фамилия студента\n" );

  // закроем файл
  fclose( students );

  return 0;        // успешное завершение процедуры
}


/* ****************************************************************************
* поиск данных одного студента в файле students.db
* ****************************************************************************/
static int find( void )
{
  // шифр студенческой группы в файле
  char grp_code[ GROUP_CODE_LEN ];

  // фамилия студента в файле
  char last_name[ NAME_LEN ];

  int found = 0;            // признак того, что запись в файле найдена
 
  // файл для хранения списков студентов
  FILE *students;

  // откроем файл в режиме чтения
  students = fopen( "students.db", "rb" );

  // проверяем успешность открытия файла
  if ( students == NULL )
  {
    fprintf( stderr, "Не могу открыть файл students.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Поиск данных студента\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  gets( grp_code );

  printf( "\nВведите фамилию студента (не более %d символов): ", NAME_LEN - 1 );
  gets( last_name );

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей описание одного студента
  while ( fread( &student, sizeof( student ), 1, students ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( student.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенные пользователем шифр группы и фамилия студента совпадают
    // с теми, которые содержатся в прочитанной записи из файла
    if ( strcmp( grp_code, student.group_code ) == 0 && 
         strcmp( last_name, student.last_name ) == 0
       )
    {
      found = 1;    // запись найдена -- присвоим флагу ненулевое значение

      // выведем найденную запись одной строкой
      printf( "%s %s %s %d %.2f\n", student.group_code, student.last_name,
                                    student.first_name, student.height,
                                    student.weight );

      break;        // выходим из цикла, т. к. запись найдена и обработана
    }
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенные пользователем ключевые данные в файле не найдены
    printf( "Неверный шифр группы или фамилия студента\n" );

  // закроем файл
  fclose( students );

  return 0;        // успешное завершение процедуры
}


/* ****************************************************************************
* вывод списка одной студенческой группы
* ****************************************************************************/
static int list( void )
{
  int written;     // счетчик числа блоков данных, успешно записанных в файл
  int i;           // счетчик числа выведенных записей

  int found = 0;   // признак того, что запись в файле найдена

  // шифр студенческой группы в файле
  char grp_code[ GROUP_CODE_LEN ];

  // файл для хранения списков студентов
  FILE *students;

  // откроем теперь файл в режиме чтения и выведем содержимое файла
  students = fopen( "students.db", "rb" );

  // проверяем успешность открытия файла
  if ( students == NULL )
  {
    fprintf( stderr, "Не могу открыть файл students.db\n" );
    return 1;      // неуспешное завершение процедуры
  }

  printf( "Списки студенческих групп\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  gets( grp_code );

  printf( "\n" );   // пустая строка на экране

  printf( "Список группы %s\n\n", grp_code );

  i = 1;   // будем нумеровать выводимые записи о студентах

  // будем в цикле читать из файла students.db структуру за структурой
  // (т. е. запись за записью)
  while ( fread( &student, sizeof( student ), 1, students ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( student.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенный пользователем шифр группы совпадает с тем, который
    // содержится в прочитанной записи из файла
    if ( strcmp( grp_code, student.group_code ) == 0 )
    {
      found = 1;    // запись найдена -- присвоим флагу ненулевое значение

      // выведем найденную запись одной строкой
      // значение поля student.group_code выводить не будем -- нет смысла
      printf( "%d. %s %s %d %.2f\n", i,
                                     student.last_name,
                                     student.first_name,
                                     student.height,
                                     student.weight );
      i++;
    }
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенный пользователем шифр группы в файле не найден
    printf( "Неверный шифр группы\n" );

  // закроем файл
  fclose( students );

  return 0;
}


/* ****************************************************************************
* корректировка данных для одного студента в файле students.db
* ****************************************************************************/
static int update( void )
{
  int  written;    // счетчик числа блоков данных, успешно записанных в файл

  char  grp_code[ GROUP_CODE_LEN ];    // шифр студенческой группы
  char  last_name[ NAME_LEN ];         // фамилия
  char  first_name[ NAME_LEN ];        // имя
  char  height[ 5 ];                   // рост (в сантиметрах)
  char  weight[ 7 ];                   // вес (в килограммах)

  int found = 0;            // признак того, что запись в файле найдена
  int updated = 0;          // признак того, что запись в файле обновлена

  // файл для хранения списков студентов
  FILE *students;

  // откроем файл в режиме чтения и записи 
  students = fopen( "students.db", "r+b" );

  // проверяем успешность открытия файла
  if ( students == NULL )
  {
    fprintf( stderr, "Не могу открыть файл students.db\n" );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  printf( "Корректировка данных студента\n\n" );

  printf( "Введите шифр студенческой группы \n" \
          "(не более %d символов без пробелов): ", GROUP_CODE_LEN - 1 );
  gets( grp_code );

  printf( "\nВведите фамилию студента (не более %d символов): ", NAME_LEN - 1 );
  gets( last_name );

  printf( "\n" );   // пустая строка на экране

  // читаем файл по одной записи, составляющей описание одного студента
  while ( fread( &student, sizeof( student ), 1, students ) == 1 )
  {
    // если эта запись помечена для удаления
    if ( student.deleted )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    // если введенные пользователем шифр группы или фамилия студента
    // НЕ совпадают с теми, которые содержатся в прочитанной записи из файла
    // (символы "||" означают логическое "или")
    if ( strcmp( grp_code, student.group_code ) != 0 || 
         strcmp( last_name, student.last_name ) != 0 )
      continue;    // переходим к началу цикла, т. е. читаем следующую запись

    found = 1;     // запись найдена -- присвоим флагу ненулевое значение

    // выведем на экран текущие значения полей записи одной строкой
    printf( "%s %s %s %d %.2f\n\n", student.group_code, student.last_name,
                                    student.first_name, student.height,
                                    student.weight );

    // попросим пользователя ввести данные
    printf( "Введите новые значения полей.\n" \
            "Если необходимо оставить прежнее значение поля, " \
            "просто нажмите клавишу Enter\n\n");

    // будем вводить каждое поле на отдельной строке
    printf( "Шифр группы (не более %d символов без пробелов): ",
            GROUP_CODE_LEN - 1 );
    gets( grp_code );

    printf( "\nФамилия (не более %d символов): ", NAME_LEN - 1 );
    gets( last_name );

    printf( "\nИмя (не более %d символов): ", NAME_LEN - 1 );
    gets( first_name );

    printf( "\nРост (в сантиметрах): " );
    gets( height );

    printf( "\nВес (в килограммах) " \
            "(десятичную часть отделяйте символом \".\"): " );
    gets( weight );

    printf( "\n" );   // пустая строка на экране

    // если ввели новое значение в поле "Шифр группы"
    if ( strlen( grp_code ) != 0 )
    {
      // скопируем новое значение в структуру 
      strcpy( student.group_code, grp_code );
      updated = 1;                       // признак -- запись обновлена
    }

    // если ввели новое значение в поле "Фамилия"
    if ( strlen( last_name ) != 0 )
    {
      // скопируем новое значение в структуру 
      strcpy( student.last_name, last_name );
      updated = 1;                        // признак -- запись обновлена
    }

    // если ввели новое значение в поле "Имя"
    if ( strlen( first_name ) != 0 )
    {
      // скопируем новое значение в структуру 
      strcpy( student.first_name, first_name );
      updated = 1;                        // признак -- запись обновлена
    }

    // если ввели новое значение в поле "Рост"
    if ( strlen( height ) != 0 )
    {
      // скопируем новое значение в структуру 
      // функция sscanf() преобразует строку в число
      sscanf( height, "%d", &student.height );
      updated = 1;                        // признак -- запись обновлена
    }

    // если ввели новое значение в поле "Вес"
    if ( strlen( weight ) != 0 )
    {
      // скопируем новое значение в структуру 
      sscanf( weight, "%f", &student.weight );
      updated = 1;                        // признак -- запись обновлена
    }

    // если было введено новое значение хотя бы для одного из полей
    if ( updated )   // это условие можно записать так: if ( updated == 1 )
    {
      // переместим указатель позиции в файле к началу этой записи (структуры)
      // параметр SEEK_CUR означает, что перемещение указателя позиции
      // выполняется от ТЕКУЩЕЙ позиции в файле; знак "-" означает перемещение
      // в сторону НАЧАЛА файла
      fseek( students, -( sizeof( student ) ), SEEK_CUR );

      // запишем введенные данные в файл students.db
      written = fwrite( &student, sizeof( student ), 1, students );

      if ( written == 1 )
        printf( "Запись обновлена\n" );
      else
        printf( "Ошибка при записи данных в файл students.db\n" );
    }
    else         // т. е. updated == 0
    {
      printf( "Запись не обновлена, т. к. не введены новые значения полей\n" );
    }
    
    break;      // выходим из цикла, т. к. запись найдена и обработана
  }

  // эта запись равносильна такой: if ( found == 0 )
  if ( !found )  // введенные пользователем ключевые данные в файле не найдены
    printf( "Неверный шифр группы или фамилия студента\n" );

  // закроем файл
  fclose( students );

  return 0;        // успешное завершение процедуры
}
