/* ****************************************************************************
* Программа: динамическое распределение памяти (часть 2)
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// этот заголовочный файл нужен для функций malloc() и free()
#include <stdlib.h>


int main( void )
{
  int i, j;    // счетчики цикла

  // В сложных программах часто используются двухмерные массивы, распределенные
  // динамически. В таких массивах, упрощенно говоря, первый уровень
  // (первое измерение) представляет собой массив указателей, а полезные данные
  // содержатся в одномерных массивах второго уровня. Пусть нам нужен
  // двухмерный массив целых чисел. Для создания такого массива мы сначала
  // должны запросить у операционной системы память для массива указателей
  // на целый тип. Затем в цикле мы будем запрашивать память уже для массивов
  // целых чисел и адреса этих массивов будем записывать в первый массив.
  // В объявлении переменной используем две звездочки **. Это объясняется тем,
  // что данная переменная должна быть указателем (отсюда возникает первая
  // звездочка), причем этот указатель будет хранить адрес динамически 
  // распределенного блока памяти, элементами которого также являются указатели
  // (отсюда возникает вторая звездочка).
  int **numbers;

  // указатель для временного хранения адреса блока памяти, распределенного для
  // одномерного массива целых чисел
  int *tmp;

  // Выполним непосредственый запрос на динамическое распределение памяти
  // для 5 указателей на целый тип. Поскольку динамический массив будет хранить
  // адреса (указатели), то в операции sizeof используется ( int * ), что 
  // означает указатель на int. В приведении типа также видим две звездочки --
  // ( int **).
  numbers = ( int **) malloc( sizeof( int * ) * 5 );

  // проверим успешность работы функции malloc()
  if ( numbers == NULL )
  {
    puts( "Не могу выделить память для массива указателей на целый тип" );
    exit( 1 );     // "аварийное" завершение программы
  }

  // теперь распределим память для массивов из 10 целых чисел и адреса этих 
  // массивов запишем в первый массив
  for ( i = 0; i < 5; i++ )
  {
    tmp = ( int * ) malloc( sizeof( int ) * 10 );

    // проверим успешность работы функции malloc()
    if ( tmp == NULL )
    {
      puts( "Не могу выделить память для массива целых чисел" );
      exit( 1 );     // "аварийное" завершение программы
    }

    // запишем адрес полученного блока памяти в массив первого уровня
    // (измерения)
    numbers[ i ] = tmp;
  }
  
  // инициализируем генератор случайных чисел
  srand( ( unsigned int )time( 0 ) );

  // заполним массивы случайными числами
  for ( i = 0; i < 5; i++ )
    for ( j = 0; j < 10; j++ )
      numbers[ i ][ j ] = rand();


  // выведем содержимое двухмерного массива
  for ( i = 0; i < 5; i++ )
  {
    for ( j = 0; j < 10; j++ )
    {
      // для выравнивания колонок чисел указываем ширину поля
      // в спецификаторе -- %5d
      printf( "%5d ", numbers[ i ][ j ] );
    }
    printf( "\n" );  // каждый массив на отдельной строке
  }

  // сначала освободим память, занимаемую массивами второго уровня
  for ( i = 0; i < 5; i++ )
    free( numbers[ i ] );

  // теперь освободим память, занимаемую массивом указателей первого уровня  
  free( numbers );

  return 0;
}
