/* ****************************************************************************
* Программа: структуры
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// вспомним директиву #define, которая позволяет объявить символическую 
// константу, которую мы будем использовать в тексте программы вместо
// конкретного числового значения. Если будет необходимо изменить это
// значение, то мы изменим его только в ОДНОМ месте, а во всех местах
// программы эту замену сделает препроцессор компилятора
#define  NAME_LEN  31


int main( void )
{
  // В языке C используется агрегатный тип данных, который называется  
  // СТРУКТУРОЙ. Это сложный объект в памяти, состоящий из нескольких полей,
  // которые определяет сам программист. Такие структуры используются для 
  // описания сложных объектов реального мира, которые имеют целый набор 
  // характеристик, для представления которых в памяти компьютера приходится 
  // использовать данные различных типов: char, int ... В качестве примера
  // рассмотрим структуру для описания персональных данных человека (конечно,
  // это очень упрощенный вариант таких данных).
  struct person 
  {
    char first_name[ NAME_LEN ];  // имя
    char last_name[ NAME_LEN ];   // фамилия
    int height;                   // рост (в сантиметрах)
    float weight;                 // вес (в килограммах)
  };

  // Сформировав описание, подобное приведенному выше, мы тем самым только
  // указали компилятору схему (шаблон) для создания структур, но память
  // под конкретный объект выделена не была. Для создания объекта в памяти
  // нужно теперь определить конкретный объект типа struct person. Делаем это:
  struct person student;

  // для заполнения полей структуры можно использовать инициализацию 
  // (т. е. присвоение начальных значений в момент создания структуры)
  // или запросить эти значения у пользователя

  // создадим еще одну структуру типа person и инициализируем ее поля
  struct person sportsman = { "John", "Smith", 198, 104.50 };

  // значения полей структуры student запросим у пользователя

  // для обращения к полям структуры используется операция точка "."
  // Не забывайте, что имя массива является адресом его первого элемента,
  // поэтому в функции scanf() возле имени массива не нужно ставить знак "&".
  // Однако при запросе значений других типов знак "&" нужен.

  // вот здесь опять пригодилась константа NAME_LEN (отнимаем 1 потому, что
  // нужно оставить еще один байт для завершающего строку символа '\0')
  printf( "Введите имя студента (не более %d символов): ", NAME_LEN - 1 );
  scanf( "%s", student.first_name );               // знак "&" НЕ нужен

  printf( "Введите фамилию студента (не более %d символов): ", NAME_LEN - 1 );
  scanf( "%s", student.last_name );                // знак "&" НЕ нужен

  printf( "Введите рост студента (в сантиметрах):" );
  scanf( "%d", &student.height );                  // знак "&" нужен

  printf( "Введите вес студента (в килограммах):" );
  scanf( "%f", &student.weight );                  // знак "&" нужен

  // теперь выведем содержимое обеих структур

  // для обращения к полям структуры используется операция точка "."
  printf( "\nСтудент\n\n" );
  printf( "Имя: %s\n", student.first_name );
  printf( "Фамилия: %s\n", student.last_name );
  printf( "Рост: %d\n", student.height );
  printf( "Вес: %.2f\n\n", student.weight );

  printf( "Спортсмен\n\n" );
  printf( "Имя: %s\n", sportsman.first_name );
  printf( "Фамилия: %s\n", sportsman.last_name );
  printf( "Рост: %d\n", sportsman.height );
  printf( "Вес: %.2f\n", sportsman.weight );

  // можно ввести все это одной строкой и поручить функции scanf() разобрать
  // эту строку на составляющие поля
  printf( "\nВведите все поля в одной строке, разделяя их пробелами:\n" );
  scanf( "%s %s %d %f", student.first_name, student.last_name, 
                        &student.height, &student.weight );

  // выведем эти сведения также одной строкой
  printf( "Студент: %s %s %d %.2f\n", student.first_name, student.last_name, 
                                      student.height, student.weight );

  return 0;
}
