/* ****************************************************************************
* Программа: массивы структур
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// этот заголовочный файл нужен для функции setlocale()
#include <locale.h>

// этот заголовочный файл нужен для функций malloc() и free()
#include <stdlib.h>

#define  NAME_LEN  31

/*
  Используем условную компиляцию для возможной обработки широких символов
  в том случае, когда в операционной системе установлена кодировка UTF-8.
  Команда компиляции будет такой:
  
  gcc -DWIDE -o prg13 prg13.c

  Подробно использование широких символов рассматривается в программе prg18.c.
*/
#ifdef WIDE
  #define  YES L'д'
  #define  NO  L'н'
  #define  FORMAT  "%lc"
#else
  #define  YES 'д'
  #define  NO  'н'
  #define  FORMAT  "%c"
#endif

// воспользуемся той же структурой, что и в предыдущей программе
struct person 
{
  char first_name[ NAME_LEN ];  // имя
  char last_name[ NAME_LEN ];   // фамилия
  int height;                   // рост (в сантиметрах)
  float weight;                 // вес (в килограммах)
};



int main( void )
{
  int i, j;       // счетчики цикла

#ifdef WIDE
  wchar_t answer; // ответ пользователя при вводе данных
#else
  char answer;    // ответ пользователя при вводе данных
#endif

  int count;      // счетчик фактически заполненных структур

  // можно создать массив любых объектов, которые предусмотрены в языке C,
  // в том числе и массив структур
  
  // массив структур можно инициализировать таким образом
  struct person students[ 3 ] = { { "Иван", "Иванов", 180, 78.56 },
                                  { "Петр", "Петров", 160, 64.88 }, 
                                  { "Андрей", "Андреев", 175, 72.40 }
                                };

  // объявим уазатель на структуру типа person
  struct person *p_manager;

  // объявим еще один указатель на структуру типа person
  struct person *managers;

  // установим для нашей программы те параметры локализации, которые
  // установлены в операционной системе
  if ( !setlocale( LC_CTYPE, "" ) )
  {
    fprintf( stderr, "Can't set locale LC_CTYPE\n" );
    exit( 1 );
  }

  // запросим у операционной системы память под массив из 3 структур
  managers = ( struct person * )malloc( sizeof ( struct person ) * 3 );

  // проверим успешность работы функции malloc()
  if ( managers == NULL )
  {
    puts( "Не могу выделить память для массива структур" );
    exit( 1 );     // "аварийное" завершение программы
  }

  // попросим пользователя ввести данные
  printf( "Введите данные для менеджеров (3 записи)\n");

  for ( i = 0; i < 3; i++ )  
  {
    // будем вводить все поля одной строкой
    printf( "\nВведите все поля в одной строке, разделяя их пробелами\n" );
    printf( "(имя, фамилия, рост, вес):\n" );
    scanf( "%s %s %d %f", managers[ i ].first_name, managers[ i ].last_name, 
                          &managers[ i ].height, &managers[ i ].weight );

    count = i;      // счетчик фактически заполненных структур

    // выберем из буфера все символы, иначе при следующем вызове
    // функции scanf() она просто возьмет эти оставшиеся в буфере символы
    // и не предложит ввести новые данные
    while ( getchar() != '\n' )
      continue;

    // если заполнены еще не все структуры, ...
    if ( i < 2 )
    {
      // ... то спросим у пользователя, хочет ли он продолжить ввод

      answer = '\0';
      while ( answer != YES && answer != NO )
      {
        printf( "\nПродолжить ввод данных (д/н) ?: " );
        scanf( FORMAT, &answer );

        // выберем из буфера все символы, иначе при следующем вызове
        // функции scanf() она просто возьмет эти оставшиеся в буфере символы
        // и не предложит ввести новые данные
        while ( getchar() != '\n' )
          continue;
      }

      if ( answer == NO )     // ответ пользователя - 'н'
        break;                // выходим из цикла досрочно
    }
  }


  // выведем сведения

  // Учтем, что пользователь мог отказаться от ввода ранее, чем были
  // заполнены все структуры, поэтому в условии окончания цикла будет
  // использована переменная count.
  // Для вывода будем использовать указатель на структуру, при этом будем
  // его наращивать в заголовке цикла. В указатель p_manager запишем адрес
  // первой структуры нашего массива. Наращивать значение указателя managers
  // не рекомендуется, т. к. тогда он будет указывать уже не на начало блока
  // памяти, выделенного нам операционной системой, а на другую ячейку, но
  // для освобождения этого блока памяти с помощью функции free() нам нужен
  // адрес именно начала блока памяти.
  // Вспомните, что при увеличении значения указателя на 1 он указывает на 
  // следующий объект того типа, для которого был создан. Таким образом, 
  // наш указатель будет указывать на следующую структуру в массиве. При 
  // использовании указателя на структуру изменится синтаксис для обращения 
  // к элементмам структуры -- вместо операции "." используем операцию "->".
  // Но зато теперь нам не нужно использовать индекс в массиве managers 
  // (мы его использовали при вводе данных).

  // Обратите внимание, что в условии окончания цикла используется неравенство
  // типа "<=", а не "<"
  for ( i = 0, p_manager = managers; i <= count; i++, p_manager++ )  
  {
    // выведем эти сведения также одной строкой
    printf( "Менеджер %d: %s %s %d %.2f\n", i + 1,
                                          p_manager->first_name,
                                          p_manager->last_name, 
                                          p_manager->height, 
                                          p_manager->weight );
  }

  // освободим память
  free( managers );

  return 0;
}
