/* ****************************************************************************
* Программа: блоковый ввод/вывод данных
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// этот заголовочный файл нужен для функции setlocale()
#include <locale.h>

// для поддержки широких символов
#include <wchar.h>

#include <stdlib.h>

#define  NAME_LEN  31

/*
  Используем условную компиляцию для возможной обработки широких символов
  в том случае, когда в операционной системе установлена кодировка UTF-8.
  Команда компиляции будет такой:
  
  gcc -DWIDE -o prg14 prg14.c

  Подробно использование широких символов рассматривается в программе prg18.c.
*/
#ifdef WIDE
  #define  YES L'д'
  #define  NO  L'н'
  #define  FORMAT  "%lc"
#else
  #define  YES 'д'
  #define  NO  'н'
  #define  FORMAT  "%c"
#endif

// воспользуемся той же структурой, что и в предыдущей программе
struct person 
{
  char first_name[ NAME_LEN ];  // имя
  char last_name[ NAME_LEN ];   // фамилия
  int height;                   // рост (в сантиметрах)
  float weight;                 // вес (в килограммах)
};

// Язык C позволяет программисту определить свои собственные имена для 
// конкретных типов данных с помощью специального ключевого слова typedef.
// Это бывает удобно при работе со сложными типами данных, например, со
// структурами.
typedef struct person PERSON;


int main( void )
{
  int i, j;        // счетчики цикла

#ifdef WIDE
  wchar_t answer;  // ответ пользователя при вводе данных
#else
  char answer;     // ответ пользователя при вводе данных
#endif

  int count = 0;   // счетчик фактически заполненных структур
  int written;     // счетчик числа блоков данных, успешно записанных в файл

  // объявим переменную типа PERSON
  // как видите, теперь уже не нужно использовать ключевое слово struct
  PERSON student;

  // объявим переменные типа FILE (ранее мы так уже делали, см. prg6.c)
  FILE *out, *in;

  // установим для нашей программы те параметры локализации, которые
  // установлены в операционной системе
  if ( !setlocale( LC_CTYPE, "" ) )
  {
    fprintf( stderr, "Can't set locale LC_CTYPE\n" );
    exit( 1 );
  }

  // Откроем файл: первый параметр - имя файла, второй - режим доступа
  // к файлу:
  // a - открыть файл в режиме дополнения, т. е. для записи только в конец
  //     файла. Если файл уже сущестует, то новые данные добавляются в конец
  //     файла, в противном случае создается новый пустой файл;
  // b - означает, что файл является двоичным, т. е. при чтении файла и при
  //     записи данных в файл не происходит преобразования символа '\n'.
  //     При чтении данных из текстового файла пара символов "возврат каретки" 
  //     (код ASCII равен 13) и "новая строка" (код ASCII равен 10) заменяется
  //     одним символом '\n'. При записи данных в текстовый файл один 
  //     символ '\n' заменяется двумя символами: "новая строка" и "возврат 
  //     каретки". В операционной системе UNIX такого преобразования не
  //     происходит, поэтому текстовый и двоичный файл обрабатываются
  //     одинаково. Соответственно, ставить символ "b" в параметрах функции
  //     fopen() в системе UNIX не нужно.

  // дадим нашему файлу расширение "db", что означает database
  out = fopen( "students.db", "ab" );

  // если файл открыть (создать) не удалось, то указатель получит значение NULL
  if ( out == NULL )
  {
    printf( "Не могу создать файл students.db\n" );

    // возвратим операционной системе код ошибки (эти коды мы можем назначать
    // сами в зависмости от того, какая ошибка случилась)
    return 1;
  }

  // попросим пользователя ввести данные
  printf( "Введите данные для студентов\n");

  // в языке C есть цикл с постусловием do...while, но мы попробуем
  // обойтись циклом с предусловием, т. е. while...

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = YES;

  while ( answer == YES )
  {
    // будем вводить все поля одной строкой
    printf( "\nВведите все поля в одной строке, разделяя их пробелами\n" );
    printf( "(имя, фамилия, рост, вес):\n" );
    scanf( "%s %s %d %f", student.first_name, student.last_name, 
                          &student.height, &student.weight );

    // запишем введенные данные в файл students.db
    // параметры функции fwrite():
    // - &student - адрес блока памяти, начиная с которого нужно записывать
    //   данные в файл;
    // - sizeof( student ) - размер одного элемента записываемых данных;
    // - 1 - число записываемых элементов данных (в нашем случае всего 1, но
    //   может быть и больше, например, может быть записан сразу целый массив
    //   структур);
    // - out - дескриптор файла, в который производится запись.
    // Возвращаемое значение - число фактически записанных блоков данных.
    written = fwrite( &student, sizeof( student ), 1, out );

    if ( written != 1 )
      printf( "Ошибка пр записи данных в файл students.db\n" );
    else
      count++;      // счетчик структур, фактически записанных в файл

    // выберем из буфера все символы, иначе при следующем вызове
    // функции scanf() она просто возьмет эти оставшиеся в буфере символы
    // и не предложит ввести новые данные
    while ( getchar() != '\n' )
      continue;

    // спросим у пользователя, хочет ли он продолжить ввод, причем, будем
    // настаивать на том, чтобы пользователь для продолжения ввода ввел "д",
    // а для завершения ввода ввел не какую-нибудь другую букву, а именно "н"
    // (чтобы отказаться от продолжения ввода осмысленно, а не поспешно)

    // присвоим переменной answer такое значение, которое позволит нам войти
    // внутрь цикла
    answer = '\0';

    while ( answer != YES && answer != NO )
    {
      printf( "\nПродолжить ввод данных (д/н) ?: " );
      scanf( FORMAT, &answer );

      // выберем из буфера все символы, иначе при следующем вызове
      // функции scanf() она просто возьмет эти оставшиеся в буфере символы
      // и не предложит ввести новые данные
      while ( getchar() != '\n' )
        continue;

    }
  }

  // закроем файл
  fclose( out );

  // откроем теперь файл в режиме чтения и выведем содержимое файла
  in = fopen( "students.db", "rb" );

  // проверяем успешность открытия файла
  if ( in == NULL )
  {
    printf( "Не могу открыть файл students.db\n" );
    return 2;      // возвратим операционной системе код ошибки
  }

  printf( "\n\nСодержимое файла students.db\n\n" );

  // будем в цикле читать файл students.db

  i = 1;   // будем нумеровать выводимые записи о студентах

  // параметры функции fread():
  // - &student - адрес блока памяти, начиная с которого нужно размещать 
  //   в памяти данные, прочитанные из файла;
  // - sizeof( student ) - размер одного элемента читаемых данных;
  // - 1 - число читаемых из файла элементов данных (в нашем случае всего 1,
  //   но может быть и больше, например, из файла может быть прочитан сразу 
  //   целый массив структур);
  // - in - дескриптор файла, из которого производится чтение данных.
  // Возвращаемое значение - число фактически прочитанных блоков данных.
  while ( fread( &student, sizeof( student ), 1, in ) == 1 )
  {
    // выведем сведения также одной строкой
    printf( "Студент %d: %s %s %d %.2f\n", i,
                                           student.first_name,
                                           student.last_name, 
                                           student.height, 
                                           student.weight );
    i++;
  }

  // закроем файл
  fclose( in );

  return 0;
}
