/* ****************************************************************************
* Программа: операции с разрядами (побитовые операции)
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// прототип функции
void bytes_to_bits( const char *bytes, int count, unsigned short mode );

// компоненты режима использования функции bytes_to_bits()
#define PRINT_NUM_BYTE           1  // единица в нулевой позиции: 00000001
#define PRINT_ADDRESS_BYTE       2  // единица в первой позиции:  00000010
#define PRINT_ONE_BYTE_PER_LINE  4  // единица во второй позиции: 00000100


int main( void )
{
  unsigned char ch;
  short int short_var;
  int int_var;
  float float_var;
  double double_var;

  unsigned char str[ 80 ];

  // Инициализирующее значение можно записать и в двоичной системе,
  // но для этого требуется префикс 0b (нужно 16 нулей и единиц, 
  // т. к. тип short -- 16-разрядный)
  // ПРИМЕЧАНИЕ. Если компилятор не поддерживает двоичные константы, тогда
  // напишите так (в шестнадцатеричной системе):
  // unsigned short int val = 0xAAAA;
  unsigned short int val = 0b1010101010101010;
  unsigned short int new_val;

  unsigned short int left_val = 12345;
  unsigned short int right_val = 54321;

  // Сформируем переменную для задания режима работы функции bytes_to_bits().
  // Для комбинирования параметров используем поразрядную операцию "ИЛИ".
  unsigned short int mode = PRINT_NUM_BYTE | PRINT_ADDRESS_BYTE |
                            PRINT_ONE_BYTE_PER_LINE;
  // unsigned short int mode = PRINT_NUM_BYTE;

  printf( "В этой программе числа выводятся на экран в двоичной системе,\n" \
          "т.е. в виде нулей и единиц.\n\n" );

  printf( "ВАЖНО! Для многобайтовых типов данных байты выводятся в порядке\n" \
          "от младшего к старшему. У младшего байта -- меньший адрес\n" \
          "в памяти.\n\n" );

  /* 
     Покажем поразрядные операции
  */

  printf( "Поразрядное отрицание \"~\" (дополнение до единицы)\n" );
  printf( "Исходное значение (битовое представление)\n" );
  bytes_to_bits( ( char * )&val, sizeof( short int ), mode );
  
  new_val = ~val;
  printf( "Новое значение\n" );
  bytes_to_bits( ( char * )&new_val, sizeof( short int ), mode );

  /* *************************************************************************/

  printf( "\nИсходные значения\n" );
  printf( "Левый операнд  %d, битовое представление: ", left_val );
  bytes_to_bits( ( char * )&left_val, sizeof( short int ), mode );
  printf( "Правый операнд %d, битовое представление: ", right_val );
  bytes_to_bits( ( char * )&right_val, sizeof( short int ), mode );

  printf( "\nПоразрядное \"И\" (&)\n\n" );
  new_val = left_val & right_val;
  bytes_to_bits( ( char * )&new_val, sizeof( short int ), mode );
  printf( "\n" );

  printf( "Поразрядное \"ИЛИ\" (|)\n\n" );
  new_val = left_val | right_val;
  bytes_to_bits( ( char * )&new_val, sizeof( short int ), mode );
  printf( "\n" );

  printf( "Поразрядное \"исключающее ИЛИ\" (^)\n\n" );
  new_val = left_val ^ right_val;
  bytes_to_bits( ( char * )&new_val, sizeof( short int ), mode );
  printf( "\n" );

  /* *************************************************************************/

  while ( 1 )
  {
    printf( "Введите один символ (для выхода из цикла -- Enter): " );
    scanf( "%c", &ch );
    
    if ( ch == '\n' )
      break;
    else
    {
      // очистим буфер ввода, в котором остался символ новой строки '\n'
      scanf( "%*[^\n]" );  // удаляем все символы вплоть до символа '\n'
      scanf( "%*c" );      // удаляем символ '\n'

      bytes_to_bits( ( char * )&ch, sizeof( char ), mode );
    }
  }

  /* Для получения значений чисел от пользователя будем использовать не
     простую функцию scanf() и очистку буфера (как мы поступили выше), 
     а комбинацию функций fgets() и sscanf(). Поступим таким образом только
     для повышения разнообразия используемых функций языка C.
  */
  while ( 1 )
  {
    printf( "Введите короткое целое число (для выхода из цикла -- Enter): " );
    fgets( str, 80, stdin );

    if ( str[ 0 ] == '\n' )   // нажали клавишу Enter, не вводя ничего больше
      break;
      
    if ( sscanf( str, "%hd", &short_var ) )
      bytes_to_bits( ( char * )&short_var, sizeof( short int ), mode );
    else    // ввели не цифру (например, букву)
      break;
  }

  while ( 1 )
  {
    printf( "Введите целое число (для выхода из цикла -- Enter): " );
    fgets( str, 80, stdin );

    if ( str[ 0 ] == '\n' )   // нажали клавишу Enter, не вводя ничего больше
      break;

    if ( sscanf( str, "%d", &int_var ) )
      bytes_to_bits( ( char * )&int_var, sizeof( int ), mode );
    else     // ввели не цифру (например, букву)
      break;
  }

  while ( 1 )
  {
    printf( "Введите число типа float (для выхода из цикла -- Enter): " );
    fgets( str, 80, stdin );

    if ( str[ 0 ] == '\n' )   // нажали клавишу Enter, не вводя ничего больше
      break;

    if ( sscanf( str, "%f", &float_var ) )
      bytes_to_bits( ( char * )&float_var, sizeof( float ), mode );
    else
      break;
  }

  while ( 1 )
  {
    printf( "Введите число типа double (для выхода из цикла -- Enter): " );
    fgets( str, 80, stdin );

    if ( str[ 0 ] == '\n' )   // нажали клавишу Enter, не вводя ничего больше
      break;

    if ( sscanf( str, "%f", &double_var ) )
      bytes_to_bits( ( char * )&double_var, sizeof( double ), mode );
    else
      break;
  }

  return 0;
}


/* Функция служит для формирования битового представления числа, переданного
   ей в качестве параметра. Она выводит это число в виде нулей и единиц.
   Если переданное значение -- многобайтовое, то функция получает адрес первого
   (младшего) байта этого числа.
   
   Квалификатор const в объявлении параметра функции означает, что ту область
   памяти, на которую указывает указатель bytes, нельзя модифицировать
   внутри функции.
*/
void bytes_to_bits( const char *bytes, int bytes_count, unsigned short mode )
{
  // Параметры:
  // bytes - значение, которое нужно представить в виде нулей и единиц
  // bytes_count - число байтов в этом значении
  // mode - режим работы функции (значение формируется путем использования
  //        поразрядной операции "ИЛИ")
  // Компоненты этого режима:
  // PRINT_NUM_BYTE -- вывод на печать номера каждого байта
  // PRINT_ADDRESS_BYTE -- вывод на печать адреса каждого байта
  // PRINT_ONE_BYTE_PER_LINE -- вывод каждого байта на отдельной строке

  unsigned char ch;     // один байт из области памяти, на которую указывает
                        // указатель bytes
  unsigned short mask;  // битовый шаблон (маска), в котором только один бит
                        // имеет значение 1, а все остальные биты равны 0
  int current_byte;     // номер текущего байта 
  int pos;              // позиция единичного бита в битовой маске
  
  /* Для проверки того, как работает квалификатор const, попробуйте 
     раскомментировать следующий оператор. Попробуйте скомпилировать
     программу. Компилятор выдаст сообщение об ошибке. */
  // *bytes = 4;

  // обработаем все байты числа
  for ( current_byte = 0; current_byte < bytes_count; current_byte++ )
  {
    // возьмем значение байта по адресу, на который указывает указатель bytes
    // с учетом смещения на величину current_byte байтов
    ch = *( bytes + current_byte );

    // Битовая маска, в которой мы установим единичку в исходную позицию 7.
    // Вспользуемся операцией сдвига влево, хотя можно было и просто
    // использовать опрацию присваивания:
    // mask = 128;
    mask = 1 << 7;   // mask = 0b00000000100000000 (единица в позиции 7)

    // если число байтов больше одного, то будем выводить номер байта и адрес 
    if ( bytes_count > 1 )
    {
      // Проверяем, установлен ли бит в позиции, отвечающей за вывод
      // номера каждого байта (операция поразрядного "И").
      if ( mode & PRINT_NUM_BYTE )
        // байты нумеруем с единицы, а не с нуля, поэтому current_byte + 1
        printf( "Байт %d ", current_byte + 1 );
    }

    // если требуется выводить адрес каждого байта
    if ( mode & PRINT_ADDRESS_BYTE )
      printf( " адрес %p ", bytes + current_byte ); 

    // биты выводим на печать, начиная с левого края, т.е. с самого старшего
    // бита, поэтому будем сдвигать маску ВПРАВО, в сторону самого младшего
    // бита       
    for ( pos = 7; pos >= 0; pos-- )
    {
      // если в той позиции, в которой маска имеет единицу, в текущем байте
      // тоже стоит единица, то на печать выводим единицу, а в противном 
      // случае выводим ноль
      // ПРИМЕЧАНИЕ. В функции printf() используем так называемый условный
      //             оператор.
      // Этот оператор можно было записать короче:
      // printf( "%d", ( ch & mask ) );
      printf( "%d", ( ch & mask ) ? 1 : 0 );

      // сдвигаем единичку в нашей маске вправо на одну позицию
      mask = mask >> 1;
    }

    // если требуется каждый байт выводить на отдельной строке
    if ( mode & PRINT_ONE_BYTE_PER_LINE )
      printf( "\n" );
    else
      printf( " " );   // будем отделять один байт от другого пробелом
  }

  // если мы выводим все байты на одной строке, то в конце нужно все-таки
  // вывести символ новой строки
  if ( ! ( mode & PRINT_ONE_BYTE_PER_LINE ) )
    printf( "\n" );

  return;
}
