/* ****************************************************************************
* Программа: квалификатор (ключевое слово) const
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// для функции malloc()
#include <stdlib.h>

/* Раскомментируйте закомментированные операторы и скомпилируйте программу.
   Посмотрите, сообщения компилятора будут различаться в зависимости 
   от того, что делает данный недопустимый оператор: присваивает ли он
   новое значение константе или пытается осуществить запись нового значения
   в неизменяемую область памяти. Например:

   prg101.c:56:3: error: assignment of read-only variable 'const_ptr_to_const'
   prg101.c:57:3: error: assignment of read-only location '*const_ptr_to_const'
*/

// прототип функции
void const_func( const char *str );

int main( void )
{
  char str[] = "Символьная строка";
  const int a = 10;    // константа типа int
  char *ptr;           // указатель на тип char

  // указатель на константу (память выделим динамически)
  const char *ptr_to_const = ( char * )malloc( 40 );

  // указатель, являющийся константой
  char * const const_ptr = ( char * )malloc( 40 );

  // константа-указатель на константу
  const char * const const_ptr_to_const = ( char * )malloc( 40 );

  // указатель на строковую константу, которая хранится в статической памяти
  // и не может быть изменена ни через какой указатель
  char * ptr_to_string = "Константа-строка"; 

  // изменение константы -- НЕДОПУСТИМО
  // a = 20;

  // в этот указатель можно записать адрес другой символьной строки, т.е.
  // можно изменить значение самого указателя
  ptr = str;
  // через него можно изменить содержимое этой строки
  ptr[ 0 ] = 'A';

  // выведем содержимое строки с помощью двух указателей: оригинального и копии
  printf( "%s %s\n", ptr, str );

  // изменение символьной строки через данный указатель -- НЕДОПУСТИМО
  // ptr_to_const[ 4 ] = 'A';

  // записать в этот указатель адрес другой символьной строки -- допустимо
  // (но сначала освободим тот блок памяти, который нам дала функция malloc())
  free( ( void * )ptr_to_const );  // выполним приведение к типу ( void * )
  ptr_to_const = str;
  printf( "%s\n", ptr_to_const );

  // записать в этот указатель адрес другой символьной строки -- НЕДОПУСТИМО
  // (поскольку сама эта переменная -- указатель -- является константой,
  // записать в нее другое значение невозможно)
  // const_ptr = str;
  
  // изменить содержимое символьной строки через этот указатель -- допустимо,
  // поскольку сам указатель не изменяется, он указывает на тот же самый
  // адрес памяти
  const_ptr[ 0 ] = 'A';

  // НЕДОПУСТИМО
  // нельзя записать новое значение адреса в этот указатель
  // const_ptr_to_const = str;

  // нельзя использовать этот указатель для изменения области памяти,
  // на которую он указывает
  // const_ptr_to_const[ 0 ] = 'A';

  // Изменить значение строки в статической памяти -- НЕДОПУСТИМО.
  // Эта ошибка проявляется не при компиляции программы, а уже при ее
  // исполнении.
  // Выполняя программу, вы заметите сообщение об ошибке сегментирования
  // (segmentation fault). Если скомпилировать программу с поддержкой
  // отладчика (параметр компиляции -g), то при пошаговом исполнении программы
  // сбой будет на этой строке.
  ptr_to_string[ 0 ] = 'A';

  // записать в указатель новое значение адреса -- допустимо
  ptr_to_string = str;

  // это функция, у которой параметром является указатель на константу
  const_func( str );

  return 0;
}

/*
  Функция, у которой параметр имеет квалификатор const. Внутри этой функнции
  нельзя изменить ту область памяти, на которую указывает такой указатель.
*/
void const_func( const char *str )
{
  printf( "%s\n", str );    // допустимый оператор

  // Раскомментируйте этот оператор и попробуйте скомпилировать программу.
  // Какое сообщение об ошибке он выдаст?
  // str[ 0 ] = 'A';         // недопустимый оператор

  return;
}
