/* ****************************************************************************
* Программа: использование указателей на функции
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* прототипы функций */
/* Если бы мы поместили определения ("тела") функций сюда, т.е. перед
   функцией main(), в которой эти функции используются, то прототипы можно
   было бы не использовать. Важно, чтобы компилятор знал параметры и тип
   возвращаемого функцией значения до первого использования этой функции.
*/
void get_word_a( char *s );
void get_word_b( char *s );
void get_word_c( char *s );

int main( void )
{
  /* для упрощения дальнейшей работы определим свой тип данных как указатель
     на функцию, которая:
     -- получает один параметр, который имеет тип char * (указатель на тип
     char);
     -- не возвращает никакого значения (поэтому -- void)
  */
  typedef void ( *f_ptr)( char * );
  
  // теперь можем объявить массив из трех указателей на функцию
  f_ptr func_array[ 3 ];

  // если бы мы не использовали оператор typedef, то объявление массива
  // выглядело бы так
  // void ( *func_array[ 3 ] )( char * );

  int i;                        // индекс в массиве
  char ch;                      // символ
  char *str;                    // указатель для выделения памяти для строки

  str = ( char * )malloc( 40 ); // выделяем память для строки символов

  // проверяем успешность операции выделения памяти
  if ( str == NULL )
  {
    fprintf( stderr, "Can't allocate memory\n" );
    exit( 1 );
  }

  // запишем в массив адреса наших функций; адресом функции является ее имя
  func_array[ 0 ] = get_word_a; 
  func_array[ 1 ] = get_word_b; 
  func_array[ 2 ] = get_word_c; 

  // попросим пользователя ввести одну из первых трех букв английского
  // алфавита, при этом не будем делать проверку введенного значения
  // для упрощения программы
  printf( "Enter an English letter, please (a, b, c): " ); 
  scanf( "%c", &ch );

  /* индекс в массиве вычисляем путем вычитания кода буквы 'a' из кода того
     символа, который введен пользователем: если пользователь ввел букву 'a',
     то получим значение индекса 0, если он ввел 'b', получим 1, если он
     ввел 'c', получим 2 */
  i = ch - 'a'; 

  // обращаемся к массиву за адресом нужной нам функции, передаем ей 
  // и параметр -- указатель на строку символов
  ( *func_array[ i ] )( str );

  // можно было вызвать нужную функцию и таким образом, т.е. не используя
  // операцию разыменования *, т.е. получения значения по такому-то адресу
  // func_array[ i ]( str );

  // печатаем слово, начинающееся с выбранной буквы
  printf( "Here is the word: %s\n", str );

  return 0;
}

// каждая из этих функций помещает слово в область памяти, выделенную
// в функции main()
void get_word_a( char *s )
{
  // первый параметр -- целевая область памяти, второй параметр -- копируемая
  // строка символов
  strcpy( s, "Apple" ); 
}
 
void get_word_b( char *s )
{
  strcpy( s, "Banana" ); 
}

void get_word_c( char *s )
{
  strcpy( s, "Carrot" ); 
}
