/* ****************************************************************************
* Программа: использование широких символов и стандарт Unicode
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <wchar.h>
#include <stdio.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>

/*
   Коды символов русского алфавита в стандарте Unicode  можно найти по адресу:
   http://www.unicode.org/charts/PDF/U0400.pdf
*/   

/* В этой программе для получения данных от пользователя используются
   функции scanf() и wscanf(). Их параллельное использование невозможно.
   Для выбора той или иной функции используется условная компиляция. Для
   выбора функции wscanf() нужно включить параметр WSCANF в команду
   компиляции:
   
   gcc -DWSCANF -o prg18 prg18.c
*/

/* В качестве эксперимента запустите программу так: 

   LANG=ru_RU.KOI8-R ./prg
   
   и обратите внимание на то, что в ряде случаев вместо букв русского
   алфавита выводятся вопросительные знаки.
*/

// прототипы функций
char *check_orientation( FILE *stream );
void print_bytes( FILE *stream, const char *str, int count );

int main( void )
{
  wchar_t wc = L'Ф';                    // один широкий символ
  wchar_t wstr[] = L"Символы wchar_t";  // массив широких символов

  // если у нас кодировка символов UTF-8, то нельзя присвоить этой переменной
  // значение, например, 'Ф', т.к. русские буквы в UTF-8 кодируются двумя
  // байтами
  char ch = 'D';                        // обычный символ (байт)

  // а записывать в строку символов типа char русские символы можно, лишь бы
  // хватило длины массива, которую мы указываем в квадратных скобках
  // (если мы не указываем длину в квадратных скобках, то компилятор сам 
  // выделит столько, сколько нужно)
  char str[] = "Символы char";          // массив байтов ("узких" символов)

  char conv[ 256 ];      // массив для конвертирования широкого символа
                         // и строки широких символов в многобайтовое 
                         // представление
  int num_bytes;         // число байтов в многобайтовом представлении символа

  // установим для нашей программы те параметры локализации, которые
  // установлены в операционной системе
  if ( !setlocale( LC_CTYPE, "" ) )
  {
    fprintf( stderr, "Can't set locale LC_CTYPE\n" );
    exit( 1 );
  }

/*#if defined( __STDC_ISO_10646__ )
    printf( "__STDC_ISO_10646__\n" );
  #endif
*/
  // Каждый поток может иметь ориентацию на вывод широких символов (wide
  // characters) или на вывод байтов.

  // определим текущую ориентацию потоков stdin, stdout и stderr
  printf( "Текущая ориентация stdout: %s\n", check_orientation( stdout ) );
  printf( "Текущая ориентация stdin: %s\n", check_orientation( stdin ) );
  printf( "Текущая ориентация stderr: %s\n", check_orientation( stderr ) );

  printf( "Число байтов в широком символе wchar_t = %d\n", sizeof( wchar_t ) );

  printf( "\nВывод широких и обычных (или многобайтных) "
          "символов в поток с ориентацией на широкие символы\n" );  

  // на широкие символы сориентируем поток stderr (этот поток также идет
  // на экран)
  if ( fwide( stderr, 1 ) != 1 )
    printf( "Не получилось назначить ориентацию stderr на широкие символы\n" );

  printf( "Текущая ориентация stderr: %s\n", check_orientation( stderr ) );

  fwprintf( stderr, L"Широкий символ: %lc\n", wc );
  fwprintf( stderr, L"Строка широких символов: %ls\n", wstr );
  fwprintf( stderr, L"Длина этой строки: широких символов - %d\n",
            wcslen( wstr ) );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию wcslen(), а не strlen(),
  // умножаем число широких символов на число байтов в широком символе)
  print_bytes( stdout, ( char * )wstr, wcslen( wstr ) * sizeof( wchar_t ) );
  
  fwprintf( stderr, L"Обычный символ: %c\n", ch );
  fwprintf( stderr, L"Строка обычных символов: %s\n", str );
  fwprintf( stderr, L"Длина этой строки: байтов - %d\n", strlen( str ) );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию strlen())
  print_bytes( stdout, str, strlen( str ) );

  // конвертируем широкий символ в многобайтовую строку и выведем ее
  num_bytes = wctomb( conv, wc );
  conv[ num_bytes ] = '\0';  // чтобы функция fwprintf знала, где конец строки
  fwprintf( stderr, L"Конвертированный широкий символ: %s\n", conv );

  // конвертируем строку широких символов в многобайтовую строку и выведем ее
  // (проверку возвращаемого значения не делаем для упрощения)
  if ( wcstombs( conv, wstr, 256 ) == -1 )
    printf( "Не удалось конвертировать строку широких символов\n" );
  else
    fwprintf( stderr, L"Конвертированная строка широких символов: %s\n",
              conv );
  

  printf( "\nВывод широких и обычных (или многобайтных) "
          "символов в поток с ориентацией на обычные символы\n" );

  // поток stdout уже имеет ориентацию на байты, т.к. мы уже использовали
  // функцию printf(), которая является байт-ориентированной
  // ПРИМЕЧАНИЕ. Сменить ориентацию потока нельзя без его закрытия.

  printf( "Текущая ориентация stdout: %s\n", check_orientation( stdout ) );

  fprintf( stdout, "Широкий символ: %lc\n", wc );
  fprintf( stdout, "Строка широких символов: %ls\n", wstr );

  fprintf( stdout, "Обычный символ: %c\n", ch );
  fprintf( stdout, "Строка обычных символов: %s\n", str );

  printf( "\nВвод широких и обычных (или многобайтных) "
          "символов из потока с ориентацией на обычные символы\n" );

  printf( "Текущая ориентация stdin: %s\n", check_orientation( stdin ) );

#ifdef WSCANF
  printf( "\nИспользуем функцию wscanf()\n" );
  printf( "\nШирокие символы\n" );

  printf( "Введите слово: " );
  wscanf( L"%ls", wstr );
  wscanf( L"%*[^\n]" );       // очистим буфер ввода,
  wscanf( L"%*c" );           // в котором остался символ новой строки
  fprintf( stdout, "Строка широких символов: %ls\n", wstr );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию wcslen(), а не strlen(),
  // умножаем число широких символов на число байтов в широком символе)
  print_bytes( stdout, ( char * )wstr, wcslen( wstr ) * sizeof( wchar_t ) );

  printf( "Введите букву: " );
  wscanf( L"%lc", &wc );
  fprintf( stdout, "Широкий символ: %lc\n", wc );

  printf( "\nОбычные символы\n" );

  printf( "Введите слово: " );
  wscanf( L"%s", str );
  wscanf( L"%*[^\n]" );       // очистим буфер ввода,
  wscanf( L"%*c" );           // в котором остался символ новой строки
  fprintf( stdout, "Строка обычных символов: %s\n", str );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию strlen())
  print_bytes( stdout, str, strlen( str ) );

  printf( "Введите латинскую букву: " );
  wscanf( L"%c", &ch );
  fprintf( stdout, "Обычный (однобайтовый) символ: %c\n", ch );
#else
  printf( "\nИспользуем функцию scanf()\n" );
  printf( "\nШирокие символы\n" );

  printf( "Введите слово: " );
  scanf( "%ls", wstr );
  scanf( "%*[^\n]" );       // очистим буфер ввода,
  scanf( "%*c" );           // в котором остался символ новой строки
  fprintf( stdout, "Строка широких символов: %ls\n", wstr );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию wcslen(), а не strlen(),
  // умножаем число широких символов на число байтов в широком символе)
  print_bytes( stdout, ( char * )wstr, wcslen( wstr ) * sizeof( wchar_t ) );

  printf( "Введите букву: " );
  scanf( "%lc", &wc );
  fprintf( stdout, "Широкий символ: %lc\n", wc );

  printf( "\nОбычные символы\n" );

  printf( "Введите слово: " );
  scanf( "%s", str );
  scanf( "%*[^\n]" );       // очистим буфер ввода,
  scanf( "%*c" );           // в котором остался символ новой строки
  fprintf( stdout, "Строка обычных символов: %s\n", str );

  // выведем шестнадцатеричное представление байтов этой строки
  // (для определения числа байтов используем функцию strlen())
  print_bytes( stdout, str, strlen( str ) );

  printf( "Введите латинскую букву: " );
  scanf( "%c", &ch );
  fprintf( stdout, "Обычный (однобайтовый) символ: %c\n", ch );
#endif

  return 0;
}


/*
  Функция для определения ориентации потока: широкие символы или байты.
*/

char *check_orientation( FILE *stream )
{
  // параметр stream -- указатель потока (stdin, stdout, stderr)

  int stream_orientation;

  /* Для того чтобы можно было в операторе return возвратить строку str,
     нужно использовать ключевое слово static, без которого эта переменная
     была бы автоматической. А поскольку автоматические переменные создаются
     в области стека, то возвращать их адреса из функций не рекомендуется.
     Попробуйте удалить ключевое слово static и перекомпилировать программу --
     компилятор выдаст предупреждение.
  */ 
  static char str[ 80 ];
  
  // определим текущую ориентацию потока
  stream_orientation = fwide( stream, 0 );

  // запишем текстовое сообщение в массив
  if ( stream_orientation > 0 )
    sprintf( str, "широкие символы (%d)", stream_orientation );
  else if ( stream_orientation < 0 )
    sprintf( str, "байты (%d)", stream_orientation );
  else
    sprintf( str, "не определена (%d)", stream_orientation );

  return str;
}

/*
  Вывод содержимого байтов символьной строки в виде шестнадцатеричных цифр.
*/
void print_bytes( FILE *stream, const char *str, int count )
{
  // параметры:
  // stream -- указатель потока (stdout, stderr)
  // str -- указатель на строку символов
  // count -- число символов (длина строки байтов)

  int i;
  
  for ( i = 0; i < count; i++ )
  {
    /* Если не включить приведение к типу unsigned char, то русские буквы
       будут представляться четырьмя байтами (по числу байтов в типе int).
       поскольку при приведении к типу unsigned int единица, если она есть
       в позиции 7 данного байта, будет растиражирована во все три байта,
       дополняющие char до int. Эти три байта будут иметь значение FF. 
    */ 
    // спецификация %02X означает: вывод шестнадцатеричных чисел, две позиции,
    // дополнить нулем слева, если цифра будет всего одна
    fprintf( stream, "(%3d: %02X) ", i,
             ( unsigned int )( unsigned char )*( str + i ) );
  }
  fprintf( stream, "\n" );

  return;
}
