/* ****************************************************************************
* Программа: массивы и циклы
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

/* Это заголовочный файл */
#include <stdio.h>

/* #define - это так называемые директивы препроцессора. С их помощью можно
   определить символические (именованные) константы. Эти константы 
   обрабатываются препроцессором до начала компиляции. Вместо них подставляются
   их значения, определенные директивами #define. Такие константы помогают
   уменьшить объем рутинной работы, связанной с возможными изменениями 
   значений тех или иных числовых констант в тексте программы.
*/
#define LEN_OF_ARRAY      5
#define FIRST_DIMENSION   3
#define SECOND_DIMENSION  4


int main( void )
{
  /* объявим переменные */

  // Одномерный массив целых чисел; его можно инициализировать, т. е. присвоить
  // начальные значения его элементам. В квадратных скобках указывается длина 
  // массива, т. е. число элементов в нем. Причем, вместо числовой константы
  // используется именованная константа. В фигурных скобках перечисляются
  // значения, которые будут присвоены элементам массива.
  int array[ LEN_OF_ARRAY ] = { 111, 222, 333, 444, 555 };

  // Двухмерный массив целых чисел; его также можно инициализировать.
  int array2[ FIRST_DIMENSION ][ SECOND_DIMENSION ] = 
            { { 11, 12, 13, 14 },
              { 21, 22, 23, 24 },
              { 31, 32, 33, 34 }
            };
  int i, j;          // счетчики циклов

  int age;           // возраст
  int units_digit;   // число единиц первого разряда в двузначном числе
  int tens_digit;    // число единиц второго разряда в двузначном числе

  /* Наша программа будет выводить целое двузначное число в словесной форме.
     Для реализации этой возможности воспользуемся двухмерными массивами,
     в которых будем хранить словесные "числовые" значения.
  */

  /* Если кодировка символов, установленная в операционной системе, является
     однобайтовой (например, KOI8-R), тогда для хранения символьных строк 
     можно использовать двухмерный массив символов; его можно инициализировать,
     т. е. присвоить начальные значения элементам. Число элементов во втором 
     измерении массива должно быть достаточным для хранения самого длинного из
     значений символьных строк. Нужно учитывать также, что каждая символьная
     строка завершается нулевым символом, который обозначается '\0'. Этот
     символ добавляет сам компилятор. Таким образом, к максимальной длине
     символьной строки нужно добавить 1. В нашей программе мы взяли длину 12.
  */ 

  /* Воспользуемся директивой условной компиляции. Она означает, что если
     НЕ определена константа препроцессора UTF8, то в процессе компиляции  
     будет участвовать фрагмент кода, размещенный между директивами
     #ifndef UTF8 и #else. Если же мы определим эту константу, тогда будет
     включен второй фрагмент кода, размещенный между директивами #else 
     и #endif. Определить константу UTF8 можно двумя способами: либо задать
     ее в начале файла директивой #define UT8, либо указать эту константу
     в команде компиляции:
     gcc -DUTF8 -o prg4 prg4.c
     Между параметрами -D и UTF8 можно ставить пробел или не ставить.
  */
#ifndef UTF8
  char tens_digits[ 8 ][ 12 ] = { "двадцать", "тридцать", "сорок",
                                  "пятьдесят", "шестьдесят", "семьдесят",
                                  "восемьдесят", "девяносто" };
#else
  /* Если в операционной системе установлена кодировка символов UTF-8,
     тогда нужно использовать второй вариант определения этого массива,
     как массива указателей на строки.
  */
  char *tens_digits[ 8 ] = { "двадцать", "тридцать", "сорок", "пятьдесят",
                             "шестьдесят", "семьдесят", "восемьдесят", 
                             "девяносто" };
#endif

  /* Для значения 0 не нужно предусматривать никакого слова. То значение,
     которое помещается в массив, называется пустой строкой - "". В конце
     каждой символьной строки помещается нулевой символ, который обозначается
     таким образом - '\0'. При том способе инициализации массива, который
     мы использовали, этот символ в конец каждой строки-инициализатора
     подставляет компилятор. Данный символ служит маркером конца строки.
     Многие библиотечные функции языка C используют этот символ для определения
     конца строки, которая подлежит обработке. Пустая строка "" состоит из
     одного только символа '\0'.
  */ 
#ifndef UTF8
  // для однобайтовой кодировки
  char units_digits[ 10 ][ 7 ] = { "", "один", "два", "три", "четыре", "пять",
                                   "шесть", "семь", "восемь", "девять" };
#else
  // для кодировки UTF-8
  char *units_digits[ 10 ] = { "", "один", "два", "три", "четыре", "пять",
                               "шесть", "семь", "восемь", "девять" };
#endif

  printf( "Выводим содержимое массива целых чисел\n\n" );

  /* Это цикл for. В его оформлении используются три выражения, которые
     отделяются друг от друга точкой с запятой. Первое выражение служит
     для инициализации счетчика цикла, оно выполняется только один раз перед
     началом выполнения цикла. Второе выражение является условием для
     выполнения очередной итерации цикла. Если это выражение истинно, то цикл
     продолжает выполняться. Если оно ложно - работа цикла прекращается.
     Третье выражение выполняется в конце каждой итерации. Как правило, в нем
     производится изменение счетчика цикла (увеличение или уменьшение).
     В нашем примере производится увеличение счетчика. Причем для этого
     используется операция инкремента, что равносильно выражению i = i + 1.
     Есть также и операция декремента. Для переменной i это будет записываться
     как i--, что равносильно выражению i = i - 1.
     Счет элементов массива начинается с нуля. В качестве индекса в массиве
     может служить переменная.
  */ 
  for ( i = 0; i < LEN_OF_ARRAY; i++ )
  {
   // выводим значение очередного элемента массива
    printf( "%d  ", array[ i ] );
  }

  printf( "\n\n" );   // пропустим пустую строку на дисплее


  printf( "Выводим содержимое двухмерного массива целых чисел\n\n" );

  /* Для вывода двухмерного массива необходимо использовать вложенный цикл.
     Внешний цикл производит обход первого измерения массива, а внутренний
     цикл обрабатывает элементы одномерного массива, находящегося на втором
     уровне.
  */
  for ( i = 0; i < FIRST_DIMENSION; i++ )
  {
    for ( j = 0; j < SECOND_DIMENSION; j++ )
    {
      printf( "%d   ", array2[ i ][ j ] );
    }
    // каждый массив второго уровня будем выводить с новой строки
    printf( "\n" );
  }

  printf( "\n\n" );


  printf( "Введите ваш возраст (целое число лет от 20 до 99): " );

  // Введем возраст как числовую переменную целого типа, а затем разобъем
  // значение на две цифры -- число единиц и число десятков.
  scanf( "%d", &age );

  // для получения числа единиц используем операцию получения остатка
  // от деления нацело
  units_digit = age % 10;

  // для получения числа десятков используем обычную операцию деления;
  // используем тот факт, что при делении целочисленных переменных результат
  // деления не округляется, а усекается до ближайшего меньшего целого числа
  tens_digit = age / 10;

  printf( "\nВаш возраст: " );

  // поскольку мы ограничили диапазон допустимых значений возраста, то индекс
  // в массиве tens_digits нам приходится уменьшать на 2. В массиве же
  // units_digits индексом служит значение числа единиц, которое мы получили,
  // поскольку счет элементов начинается с нуля и число единиц также начинается
  // с нуля.
  printf( "%s %s\n", tens_digits[ tens_digit - 2 ],
                     units_digits[ units_digit ] );

  return 0;
}
