/* ****************************************************************************
* Программа: указатели
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

/* Это заголовочный файл для функций ввода/вывода */
#include <stdio.h>

/* Этот заголовочный файл нужен для функции strlen() */
#include <string.h>

/* прототипы функций */ 
void swap( int *first, int *second );
void process_2d_arrays( void );

int main( void )
{
  int abc = 123;

  // это объявление переменной такого типа, который называется УКАЗАТЕЛЕМ 
  // на тип int;
  // можно сразу инициализировать этот указатель, записав в него адрес
  // какой-либо переменной типа int; знак & означает взятие адреса переменной
  int *p_abc = &abc;

  // счетчик в массиве
  int i;

  // объявляем и инициализируем массив целых чисел
  // (обратите внимание, что в данном случае можно не указывать число
  // элементов)
  int array[] = { 10, 20, 30, 40, 50 };

  // еще один указатель на тип int (этот указатель не инициализирован,
  // поэтому в нем хранится "мусор")
  int *p_array;

  // создадим символьную строку
  // (обратите внимание, что в данном случае можно не указывать число
  // элементов)
  char str[] = "TEST";

  // переменные для использования в функции swap()
  int first, second;

  printf( "Значение переменной abc = %d, значение указателя p_abc = %p\n",
           abc, p_abc );

  // к значению переменной abc можно обратиться и через указатель на эту
  // переменную
  printf( "\nЗначение, находящееся в ячейке, адрес которой хранится " \
          "в указателе p_abc: %d\n\n", *p_abc );

  // для определения длины массива воспользуемся операцией sizeof: число
  // элементов в массиве равно частному от деления размера массива (в байтах)
  // на размер типа данных (в байтах), элементы которого хранятся в этом
  // массиве;
  // спецификация %p означает указатель, т. е. адрес
  for ( i = 0; i < sizeof( array ) / sizeof( int ); i++ ) 
  {
    printf( "array[ %d ] = %d     адрес = %p\n", i, array[ i ], &array[ i ] );
  }

  printf( "\nРазность указателей для двух соседних элементов массива:\n" );

  // Разность указателей двух соседних элементов массива равна 1, т. к. таковы
  // правила арифметики указателей: эта разность показывает, сколько объектов
  // конкретного типа (в нашем случае - int) помещается между ячейками памяти,
  // на которые указывают эти два указателя.
  printf( "array[ 2 ] - array[ 1 ] = %p - %p = %d\n\n",
          &array[ 2 ], &array[ 1 ], &array[ 2 ] - &array[ 1 ]
        );

  // запишем в указатель адрес первого элемента массива (имя массива является
  // адресом его первого элемента, поэтому знак & не нужен)
  p_array = array;

  // покажем еще один способ обработки всего массива с использованием
  // указателей
  for ( i = 0; i < sizeof( array ) / sizeof( int ); i++ ) 
  {
    // эта запись означает следующее: взять значение по адресу, который
    // хранится в переменной p_array, умножить это значение на 2 и записать
    // его опять по адресу, который хранится в переменной p_array
    *p_array = *p_array * 2;

    // в этом операторе в качестве параметра функции printf() можно вместо
    // array[ i ] написать *p_array
    printf( "array[ %d ] = %d\n", i, array[ i ] );

    // Прибавим к указателю 1, в результате он указывает на следующий
    // элемент в массиве.
    // ВАЖНО. Поступать таким же образом с именем массива array нельзя.
    p_array++;
  }

  printf( "\n" );

  // Выведем массив символов посимвольно, используя цикл
  // Функция strlen() определяет длину символьной строки (ее завершает
  // символ '\0')
  for ( i = 0; i < strlen( str ); i++ ) 
  {
    printf( "%c    ", str[ i ] );
  }

  printf( "\n" );

  printf( "-------------------------------------------\n" );

  // Покажем использование указателей для изменения значений переменных
  // внутри функций.
  printf( "Обмен значений двух переменных\n" );
  first = 10;
  second = 20;
  printf( "Переменная first = %d, переменная second = %d\n", first, second );
  swap( &first, &second );
  printf( "Переменная first = %d, переменная second = %d\n", first, second );

  printf( "-------------------------------------------\n" );

  // Покажем способы обращения к элементам двумерного массива символов.
  process_2d_arrays();

  return 0;
}

/* 
  Функция, выполняющая обмен значений двух переменных
*/
void swap( int *first, int *second )
{
  // параметры -- адреса двух переменных типа int

  int temp;           // переменная для временного хранения числа
  
  temp = *first;      // сохраним значение первой переменной
  *first = *second;   // запишем значение второй переменной в первую
  *second = temp;     // запишем значение первой переменной во вторую

  // Выражение *first = *second можно истолковать так: записать значение,
  // хранящееся по адресу, на который указывает переменная second, 
  // в ту область памяти, на которую указывает переменная first. Операция
  // *first называется операцией разыменования указателя (dereferencing 
  // pointer), или взятия значения по адресу.
  
  return;
}

/* 
  Функция, иллюстрирующая способы обращения к элементам двумерного массива
  символов
*/
void process_2d_arrays( void )
{
  char arr[ 3 ][ 4 ] = { 'a', 'b', 'c', '\0',
                         '1', '2', '3', '\0',
                         '~', '!', '@', '\0' 
                       };
  /* замените предыдущее определение массива на это и посмотрите, 
     что получится */
  /*  char arr[ 3 ][ 4 ] = { 'a', 'b', 'c', 'd',
                         '1', '2', '3', '4',
                         '~', '!', '@', '$' 
                       };*/
  int i, j;
    
  /* выводим элементы двумерного массива по одному (не выводим символы '\0') */
  /* во внешнем цикле перебираем строки */
  printf( "\nВывод массива по одному элементу (кроме элементов '\\0')\n\n" );
  for ( i = 0; i < 3; i++ )
  {
    /* перебираем элементы в текущей строке */
    for ( j = 0; j < 3; j++ )
    {
      printf( "%c ", arr[ i ][ j ] );

      /* А таким образом можно вывести содержимое двухмерного массива,
         используя указатели. Элементами двухмерного массива arr являются 
         одномерные массивы. Поэтому операция ( arr + i ) получает указатель
         на одномерный массив (т.е. указатель на указатель на тип char),
         операция *( arr + i ) получает одномерный массив (т.е. указатель на 
         тип char), операция *( arr + i ) + j вычисляет адрес j-го элемента 
         типа char в i-м одномерном массиве, операция *( *( arr + i ) + j )
         получает значение этого j-го элемента.
      */

      /* Закомментируйте предыдущий оператор, а этот раскомментируйте --
         вывод на экран будет точно таким же. */ 
      // printf( "%c ", *( *( arr + i ) + j ) );
    }
    printf( "\n" );
  }    

  /* Можно вывести массив и таким образом, но вывод остановится на последнем
     элементе первой строки. Если заменить символы '\0' в конце первых двух
     строк на другие символы, то будет выведен весь двухмерный массив в виде
     одной строки.
  */
  printf( "\nВесь массив arr = %s\n\n", arr );      

  /* Компилятор знает, как найти начало каждой строки массива. Поскольку
     размеры массива (число строк и столбцов) указаны в его объявлении,
     следовательно, прибавляя к адресу начала массива номер строки,
     компилятор на самом деле учитывает число элементов в строке и размер
     элементов, т.е. умножает номер строки на число элементов в ней и на
     размер одного элемента.
  */
  printf( "Вывод строк массива\n\n" );
  for ( i = 0; i < 3; i++ )
  {
    printf( "Массив типа char: arr[ %d ] = %s\n", i, arr[ i ] );
    printf( "Массив типа char: *( arr + %d ) = %s\n", i, *( arr + i ) );
    printf( "Указатель на тип char: &arr[ %d ][ 0 ] = %s\n", i, 
                                                             &arr[ i ][ 0 ] );
    printf( "Указатель на массив типа char: arr + %d = %s\n", i, arr + i );
    printf( "\n" );
  }

  printf( "Вывод адресов первых элементов каждой строки\n\n" );
  for ( i = 0; i < 3; i++ )
  {
    printf( "Адрес arr + %d = %p\n", i, arr + i );
    printf( "Адрес &arr[ %d ] = %p\n", i, &arr[ i ] );
    printf( "Адрес *( arr + %d ) = %p\n", i, *( arr + i ) );
    printf( "Адрес &arr[ %d ][ 0 ] = %p\n", i, &arr[ i ][ 0 ] );
    printf( "\n" );
  }

  printf( "Адрес массива arr = %p\n", arr );

  printf( "Обратите внимание на адреса arr, &arr[ 0 ] и &arr[ 0 ][ 0 ]\n" );

  return;
}
