/* ****************************************************************************
* Программа: файлы и циклы
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

/* Это заголовочный файл */
#include <stdio.h>

/* Директива препроцессора позволяет указать компилятору, что данный
   фрагмент кода нужно включить в обработку, если НЕ определена константа UNIX.
   Если вы компилируете этот файл в среде Linux (FreeBSD), тогда нужно эту
   константу определить, чтобы компилятор не пытался найти файл conio.h.
   Ее можно определить в этом же файле директивой 
   
   #define UNIX
   
   или указать в качестве параметра компилятора
   
   gcc -DUNIX -o prg6 prg6.c
*/
#ifndef UNIX
// этот заголовочный файл нужен для использования функции getche()
// (это не работает в среде UNIX)
#include <conio.h>
#endif

int main( void )
{
  /* объявим переменные */

  FILE *out;   // указатель на дескриптор файла для записи данных
  FILE *in;    // указатель на дескриптор файла для чтения данных

  // символьный массив для получения ввода от пользователя
  char str[ 80 ];

  // переменная для получения ответа пользователя; инициализируем ее нулем,
  // чтобы первая проверка условия цикла прошла успешно
  int user_answer = 0;

  // Создадим новый файл, а если он уже существует, он будет перезаписан.
  // Первый параметр - имя файла, второй - режим доступа к файлу. Существуют
  // следующие режимы:
  // r - открыть существующий файл только для чтения;
  // w - открыть файл только для записи. Если файл уже сущестует, то он
  //     обрезается до нулевой длины, в противном случае создается новый файл;
  // a - открыть файл в режиме дополнения, т. е. для записи только в конец
  //     файла. Если файл уже сущестует, то новые данные добавляются в конец
  //     файла, в противном случае создается новый пустой файл;
  // r+ - открыть существующий файл для чтения и записи. Исходное содержимое
  //     файла не изменяется, а начальная позиция указателя в файле - это
  //     начало файла;
  // w+ - открыть файл для чтения и записи. Если файл уже сущестует, то он
  //     обрезается до нулевой длины, в противном случае создается новый файл;
  // a+ - открыть или создать файл для чтения и дополнения в конец файла.
  //     Если файл уже сущестует, то его исходное содержимое не изменяется,
  //     в противном случае создается новый файл. Начальная позиция для чтения
  //     из файла - в начале файла, но добавление новых данных производится
  //     всегда только в конец файла.
  out = fopen( "test.txt", "w" );

  // если файл открыть (создать) не удалось, то указатель получит значение NULL
  if ( out == NULL )
  {
    printf( "Не могу создать файл test.txt\n" );

    // возвратим операционной системе код ошибки (эти коды мы можем назначать
    // сами в зависмости от того, какая ошибка случилась)
    return 1;
  }


  /* Рассмотрим цикл с предусловием */  
  while ( user_answer != 'q' )     // пока не нажата клавиша 'q'
  {
    printf( "Введите вашу фамилию: ");

    // эта функция требует для работы наличия буфера, в который будут
    // помещаться прочитанные данные
    gets( str );

    // будем записывать (выводить) данные в файл
    // функция принимает три или более параметров: дескриптор потока (файла),
    // строку формата, список переменных
    fprintf( out, "Фамилия: %s\n", str );

    printf( "Введите ваше имя: ");
    gets( str );
    fprintf( out, "Имя: %s\n", str );

    printf( "Введите ваше отчество: ");
    gets( str );
    fprintf( out, "Отчество: %s\n", str );

    // для "склеивания" строковых констант используем символ "\" в конце
    // строки исходного текста
    printf( "Для завершения ввода данных введите q, " \
            "для продолжения - любую другую клавишу: ");
#ifndef UNIX
    user_answer = getche();
#else
    // в среде UNIX можно использовать такую функцию, но в этом случае 
    // потребуется еще нажимать клавишу Enter
    user_answer = getc( stdin );
#endif
  }

  // закроем файл
  fclose( out );

  // откроем теперь файл в режиме чтения и выведем содержимое файла
  in = fopen( "test.txt", "r" );

  // проверяем успешность открытия файла
  if ( in == NULL )
  {
    printf( "Не могу открыть файл test.txt\n" );
    return 2;      // возвратим операционной системе код ошибки
  }

  printf( "\n\nСодержимое файла test.txt\n\n" );

  // Параметры функции fgets(): массив для размещения в нем прочитанной
  // строки файла, максимальная длина строки, указатель на открытый файл.
  // Если достигнут конец файла, то значением указателя будет NULL.
  while ( fgets( str, 80, in ) != NULL )
  {
    // stdout - это так называемое устройство стандартного вывода,
    // по умолчанию - дисплей
    fputs( str, stdout );
  }

  // закроем файл
  fclose( in );

  return 0;
}
