/* ****************************************************************************
* Программа: аргументы командной строки
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// заголовочный файл нужен для функции strlen()
#include <string.h>


// функция main() тоже может принимать параметры
int main( int argc, char *argv[] )
{
  int i, j;    // счетчики циклов

  /* При запуске программы ей можно передать параметры в командной строке.
     Количество параметров записывается в переменную argc. Значения параметров
     передаются программе в виде символьных строк, а точнее, в виде массива
     указателей на строки. При этом первым параметром считается сама программа
     (исполняемый файл). Поэтому реальное число параметров будет на 1 меньше,
     чем показывает переменная argc.
  */
  printf( "Число параметров, переданных программе: %d\n", argc - 1 );

  // если параметры в командной строке не переданы, то выведем инструкцию
  // для пользователя
  if ( argc == 1 )
  {
    // элемент массива argv[] с индексом 0 содержит имя исполняемого файла;
    // поскольку при компиляции исполняемому файлу могло быть дано какое
    // угодно имя, мы возьмем это фактическое имя из массива argv[]
    printf( "Запуск программы: %s параметр1 параметр2 параметр3 ...\n",
            argv[ 0 ] );
    return 0;
  }

  printf( "Параметры, переданные программе:\n\n" );

  // обратите внимание, что начальное значение i равно 1, а не 0,
  // т. к. элемент argv[ 0 ] содержит имя исполняемого файла
  //
  for ( i = 1; i < argc; i++ )
  {
    // выводим ЗНАЧЕНИЯ параметров, поэтому используем спецификацию %s
    printf( "%s\n", argv[ i ] );
  }

  printf( "\nАдреса параметров в памяти:\n\n" );

  for ( i = 1; i < argc; i++ )
  {
    // отличие от предыдущего цикла в том, что теперь мы используем
    // спецификацию %p, т. к. нам нужно вывести адрес (указатель);
    // массив argv[] содержит адреса символьных строк -- параметров,
    // т. е. указатели на них
    printf( "адрес параметра %d: %p\n", i, argv[ i ] );
  }

  printf( "\nАдреса элементов массива argv[]:\n\n" );

  for ( i = 1; i < argc; i++ )
  {
    // отличие от предыдущего цикла в том, что теперь мы используем
    // знак & с именем массива argv[], тем самым мы получаем не значение
    // элемента массива argv[], а адрес этого элемента;
    // заметьте, что размер каждого элемента этого массива
    // равен sizeof (char *)
    printf( "адрес элемента номер %d: %p\n", i, &argv[ i ] );
  }

  // символ "\" в конце строки позволяет "склеить" строковый литерал,
  // который выведет функция printf()
  // ВАЖНО! После этого символа "\" нужно сразу нажать клавишу Enter.
  printf( "\nПосимвольный вывод параметров в формате: "   \
          "'символ' (адрес символа)\n\n" );

  for ( i = 1; i < argc; i++ )
  {
    printf( "Параметр %d: ", i );

    // функция strlen() дает длину символьной строки (без нулевого символа,
    // являющегося завершающим элементом каждой символьной строки)
    for ( j = 0; j < strlen( argv[ i ] ); j++ )
    {
      // чтобы читать символ за символом, используем индексы уже для 
      // двухмерного массива
      printf( "'%c' (%p)  ", argv[ i ][ j ], &argv[ i ][ j ] );
    }
    printf( "\n" );     // каждый параметр выводим на отдельной строке
  }

  return 0;
}
