/* ****************************************************************************
* Программа: список студентов
* Модуль:    главный модуль
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// этот заголовочный файл нужен для функции setlocale()
#include <locale.h>

// эти заголовочные файлы нужны для функций open(), close() и chsize()
#include <fcntl.h>
//#include <io.h>

// в этом заголовочном файле определена глобальная переменная errno
// и ее значения
#include <errno.h>

// этот заголовочный файл нужен для функций malloc() и free()
#include <stdlib.h>

// в этом заголовочном файле определены структуры GROUP и STUDENT,
// которые используются в нашей функции service()
#include "students.h"


// прототип
static int service( char *filename );



/* ****************************************************************************
* главная функция программы
******************************************************************************/
int main( void )
{
  char answer;     // ответ пользователя при выборе пункта меню

  // установим для нашей программы те параметры локализации, которые
  // установлены в операционной системе
  if ( !setlocale( LC_ALL, "" ) )
  {
    fprintf( stderr, "Не могу установить параметры локализации LC_CTYPE\n" );
    exit( 1 );
  }

  // присвоим переменной answer такое значение, которое позволит нам войти
  // внутрь цикла
  answer = '1';

  while ( answer == '1' || answer == '2' || answer == '3' )
  {
    printf( "1. Работа с описаниями студенческих групп\n" );
    printf( "2. Работа со списками студенческих групп\n" );
    printf( "3. Обслуживание базы данных\n" );
    printf( "4. Выход из программы\n\n" );

    printf( "Выберите действие: " );
    scanf( "%c", &answer );      // ответ получим в виде символа, а не числа

    // очистим буфер ввода
    while ( getchar() != '\n' )
      continue;

    if ( answer == '1' )
    {
      groups();
    }
    else if ( answer == '2' )
    {
      lists();
    }
    else if ( answer == '3' )
    {
      service( "groups.db" );
      service( "students.db" );
    }
    else if ( answer == '4' )
    {
      return 0;
    }
    else    // выбран какой-то другой символ
    {
      printf( "Укажите правильный номер пункта меню\n" );
    }
  }
  return 0;
}


/* ****************************************************************************
* обслуживание базы данных
* Эта функция удаляет из файлов нашей простой базы данных те записи, которые
* имеют метку "*" (т. е. предназначены для физического удаления)
******************************************************************************/
static int service( char *filename )
{
  // параметр:
  // char *filename - имя файла данных

  // Идея алгоритма этой функции такова: рассматривать области файла, 
  // занимаемые записями, помеченными для удаления, как свободные. Сдвинуть
  // записи, не помеченные для удаления, к началу файла, используя свободные
  // области. После завершения процедуры "дефрагментации" файла выполнить
  // его усечение. Если в файле нет ни одной записи, помеченной для удаления,
  // то каждая запись будет прочитана и просто вновь записана на свое же место.

  // счетчик числа блоков данных, успешно записанных в файл
  int written;

  // позиция указателя в файле данных, начиная с которой будет продолжено
  // чтение файла на следующей итерации цикла
  // ПРИМЕЧАНИЕ. Тип данных long int - это длинное целое.
  long int position_read;

  // позиция указателя в файле данных, начиная с которой будет продолжена
  // запись данных в этот файл на следующей итерации цикла
  long int position_write;

  // дескриптор файла, который возвращает функция open() при его открытии
  int handle;

  // результат работы функции chsize(), которая укорачивает файл до нужной
  // длины (если функция завершилась успешно, то она возвращает значение 0)
  int result;

  // длина записи в файле данных (число байтов)
  int record_len;

  // указатель на область памяти, в которую будет считываться одна запись
  // из файла данных
  char *record;

  // поток для работы с файлом данных
  FILE *db;

  // для получения блока памяти, в который будет считываться запись из файла,
  // нужно знать длину этой записи

  // если функция "дефрагментирует" файл groups.db, то длина записи будет
  // равна размеру структуры GROUP, определенной в файле students.h
  if ( strcmp( filename, "groups.db" ) == 0 )
    record_len = sizeof( GROUP );

  // если функция "дефрагментирует" файл students.db -- структура STUDENTS
  else if ( strcmp( filename, "students.db" ) == 0 )
    record_len = sizeof( STUDENT );

  else
  {
    fprintf( stderr, "Неверное имя файла данных: %s\n", filename );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }  

  // запросим у операционной системы столько байтов памяти, сколько необходимо
  // для размещения одной записи из конкретного файла данных (groups.db или
  // students.db)
  record = ( char * )malloc( record_len );

  // откроем файл в режиме чтения и записи 
  db = fopen( filename, "r+b" );

  // проверяем успешность открытия файла
  if ( db == NULL )
  {
    fprintf( stderr, "Не могу открыть файл %s\n", filename );
    return 1;   // возвратим ненулевое значение, означающее наличие ошибки
  }

  // в начале процесса "дефрагментации" файла позиция, с которой будет
  // продолжена запись в файл, конечно же равна 0
  position_write = 0;

  // читаем файл по одной записи
  // параметры функции fread() следующие:
  // - прочитанные данные размещаются в область памяти с адресом record
  // (обратите внимание, что знак "&" не нужен);
  // - размер блока данных составляет 1 байт;
  // - число таких блоков равно record_len;
  // - указатель на открытый поток для работы с файлом - db
  while ( fread( record, 1, record_len, db ) == record_len )
  {
    // проверим, не помечена ли если эта запись для удаления
    // Для того, чтобы получить значение признака удаленной записи, было бы
    // удобно обратиться к полю структуры по его имени, однако мы прочитали
    // запись из файла в виде массива символов, а не в виде структуры. Выход
    // состоит в том, чтобы преобразовать адрес массива символов в указатель
    // на структуру. Поскольку мы получили указатель, то для обращения
    // к элементам структуры нужно использовать операцию "->". Конечно,
    // пришлось использовать и скобки, например, ( ( GROUP * )record )->deleted
    if ( ( strcmp( filename, "groups.db" ) == 0 && 
           ( ( GROUP * )record )->deleted 
         ) ||
         ( strcmp( filename, "students.db" ) == 0 && 
           ( ( STUDENT * )record )->deleted 
         )
       )
      // поскольку только что прочитанная запись помечена для удаления,
      // то переходим к началу цикла, т. е. читаем следующую запись
      continue;

    // нашли неудаленную запись -- нужно переместить ее ближе к началу
    // файла на свободное место
    
    // запомним текущую позицию указателя в файле, с этой позиции мы продолжим
    // ЧТЕНИЕ файла
    position_read = ftell( db );

    // переместим указатель позиции в файле к тому месту, с которого мы
    // продолжим ЗАПИСЬ в файл (параметр SEEK_SET означает, что отсчет
    // позиции выполняется от НАЧАЛА файла)
    fseek( db, position_write, SEEK_SET );
  
    // запишем данные в файл filename
    // параметры функции fwrite() следующие:
    // - в файл записываются данные, размещенные в области памяти
    // с адресом record (обратите внимание, что знак "&" не нужен);
    // - размер блока данных составляет 1 байт;
    // - число таких блоков равно record_len;
    // - указатель на открытый поток для работы с файлом - db
    written = fwrite( record, 1, record_len, db );

    // функция fwrite() возвращает число успешно записанных блоков данных
    if ( written != record_len )
    {
      fprintf( stderr, "Ошибка при записи данных в файл %s\n", filename );
      return 1;
    }

    // запомним текущую позицию указателя в файле, с этой позиции мы продолжим
    // ЗАПИСЬ в файл
    position_write = ftell( db );

    // вернемся к той позиции в файле, с которой нужно продолжить чтение данных
    fseek( db, position_read, SEEK_SET );
  }

  // закроем файл
  fclose( db );

  // Теперь мы должны укоротить файл, отрезав у него "хвост", записи
  // из которого уже перенесены ближе к началу файла. Новая длина файла
  // определяется согласно значению переменной position_write, потому что
  // после завершения обработки всего файла эта переменная указывает на
  // тот байт в файле, который расположен НЕПОСРЕДСТВЕННО за последней
  // перемещенной записью.

#ifdef UNIX
  // в операционной системе UNIX можно использовать следующий вызов функции
  // ПРИМЕЧАНИЕ. Тут надо бы сделать проверку успешности работы этой функции.
  truncate( filename, position_write );
#else
  // в операционной системе Windows поступаем следующим образом

  // откроем файл с помощью этой функнции
  // второй параметр означает, что файл открывается в режиме чтения и записи,
  // а также что он рассматривается как двоичный файл (такая форма параметра -
  // O_RDWR|O_BINARY - означает, что его значение формируется из двух
  // компонентов с помощью битовой логической операции "|", т. е. "ИЛИ")
  handle = open( filename, O_RDWR|O_BINARY);

  // укоротим файл до длины position_write
  result = chsize( handle, position_write );

  // закроем файл
  close( handle );

  // в случае ошибки функция chsize() возвратит ненулевое значение
  // (выведем код ошибки, который записывается в глобальную переменную errno)
  if ( result != 0 )   
    printf( "Код ошибки: %d\n", errno );
#endif

  // освободим память
  free( record );

  return 0;
}
